from slixmpp import register_stanza_plugin
from slixmpp.plugins.xep_0060.stanza import Item
from slixmpp.xmlstream import ElementBase

NS = "urn:xmpp:bookmarks:1"


class Conference(ElementBase):
    namespace = NS
    name = "conference"
    plugin_attrib = "conference"
    interfaces = {"name", "autojoin", "nick", "password"}
    sub_interfaces = {"nick", "password"}

    def set_autojoin(self, v: bool):
        self._set_attr("autojoin", "true" if v else "false")

    def get_autojoin(self):
        v = self._get_attr("autojoin", "")
        if not v:
            return False
        return v == "1" or v.lower() == "true"


class Extensions(ElementBase):
    namespace = NS
    name = "extensions"
    plugin_attrib = "extensions"


def register_plugin():
    register_stanza_plugin(Conference, Extensions)
    register_stanza_plugin(Item, Conference)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2011 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import logging
import hashlib
import base64

from asyncio import Future, Lock
from collections import defaultdict

from slixmpp import __version__
from slixmpp.stanza import StreamFeatures, Presence, Iq
from slixmpp.xmlstream import register_stanza_plugin, JID
from slixmpp.xmlstream.handler import Callback
from slixmpp.xmlstream.matcher import StanzaPath
from slixmpp.util import MemoryCache
from slixmpp.exceptions import XMPPError
from slixmpp.plugins import BasePlugin
from slixmpp.plugins.xep_0115 import stanza, StaticCaps
from slixmpp.types import OptJidStr


log = logging.getLogger(__name__)


class XEP_0115(BasePlugin):

    """
    XEP-0115: Entity Capabilities
    """

    name = 'xep_0115'
    description = 'XEP-0115: Entity Capabilities'
    dependencies = {'xep_0030', 'xep_0128', 'xep_0004'}
    stanza = stanza
    default_config = {
        'hash': 'sha-1',
        'caps_node': None,
        'broadcast': True,
        'cache': None,
    }

    def plugin_init(self):
        self.hashes = {'sha-1': hashlib.sha1,
                       'sha1': hashlib.sha1,
                       'md5': hashlib.md5}

        if self.caps_node is None:
            self.caps_node = 'http://slixmpp.com/ver/%s' % __version__

        if self.cache is None:
            self.cache = MemoryCache()

        register_stanza_plugin(Presence, stanza.Capabilities)
        register_stanza_plugin(StreamFeatures, stanza.Capabilities)

        self._disco_ops = ['cache_caps',
                           'get_caps',
                           'assign_verstring',
                           'get_verstring',
                           'supports',
                           'has_identity']

        self.xmpp.register_handler(
                Callback('Entity Capabilities',
                         StanzaPath('presence/caps'),
                         self._handle_caps))

        self.xmpp.add_filter('out', self._filter_add_caps)

        self.xmpp.add_event_handler('entity_caps', self._process_caps)

        if not self.xmpp.is_component:
            self.xmpp.register_feature('caps',
                    self._handle_caps_feature,
                    restart=False,
                    order=10010)

        disco = self.xmpp['xep_0030']
        self.static = StaticCaps(self.xmpp, disco.static)

        for op in self._disco_ops:
            self.api.register(getattr(self.static, op), op, default=True)

        for op in ('supports', 'has_identity'):
            self.xmpp['xep_0030'].api.register(getattr(self.static, op), op)

        self._run_node_handler = disco._run_node_handler

        disco.cache_caps = self.cache_caps
        disco.update_caps = self.update_caps
        disco.assign_verstring = self.assign_verstring
        disco.get_verstring = self.get_verstring

        # prevent concurrent fetches for the same hash
        self._locks = defaultdict(Lock)

    def plugin_end(self):
        self.xmpp['xep_0030'].del_feature(feature=stanza.Capabilities.namespace)
        self.xmpp.del_filter('out', self._filter_add_caps)
        self.xmpp.del_event_handler('entity_caps', self._process_caps)
        self.xmpp.remove_handler('Entity Capabilities')
        if not self.xmpp.is_component:
            self.xmpp.unregister_feature('caps', 10010)
        for op in ('supports', 'has_identity'):
            self.xmpp['xep_0030'].restore_defaults(op)

    def session_bind(self, jid):
        self.xmpp['xep_0030'].add_feature(stanza.Capabilities.namespace)

    async def _filter_add_caps(self, stanza):
        if not isinstance(stanza, Presence) or not self.broadcast:
            return stanza

        if stanza['type'] not in ('available', 'chat', 'away', 'dnd', 'xa'):
            return stanza

        ver = await self.get_verstring(stanza['from'])
        if ver:
            stanza['caps']['node'] = self.caps_node
            stanza['caps']['hash'] = self.hash
            stanza['caps']['ver'] = ver
        return stanza

    def _handle_caps(self, presence):
        if not self.xmpp.is_component:
            if presence['from'] == self.xmpp.boundjid:
                return
        self.xmpp.event('entity_caps', presence)

    def _handle_caps_feature(self, features):
        # We already have a method to process presence with
        # caps, so wrap things up and use that.
        p = Presence()
        p['from'] = self.xmpp.boundjid.domain
        p.append(features['caps'])
        self.xmpp.features.add('caps')

        self.xmpp.event('entity_caps', p)

    async def _process_caps(self, pres: Presence):
        if not pres['caps']['hash']:
            log.debug("Received unsupported legacy caps: %s, %s, %s",
                    pres['caps']['node'],
                    pres['caps']['ver'],
                    pres['caps']['ext'])
            self.xmpp.event('entity_caps_legacy', pres)
            return

        ver = pres['caps']['ver']
        async with self._locks[ver]:
            await self._process_caps_wrapped(pres, ver)
        self._locks.pop(ver, None)

    async def _process_caps_wrapped(self, pres: Presence, ver: str):
        existing_verstring = await self.get_verstring(pres['from'].full)
        if str(existing_verstring) == str(ver):
            return

        existing_caps = await self.get_caps(verstring=ver)
        if existing_caps is not None:
            await self.assign_verstring(pres['from'], ver)
            return

        ifrom = pres['to'] if self.xmpp.is_component else None

        if pres['caps']['hash'] not in self.hashes:
            try:
                log.debug("Unknown caps hash: %s", pres['caps']['hash'])
                await self.xmpp['xep_0030'].get_info(jid=pres['from'], ifrom=ifrom)
                return
            except XMPPError:
                return

        log.debug("New caps verification string: %s", ver)
        try:
            node = '%s#%s' % (pres['caps']['node'], ver)
            caps = await self.xmpp['xep_0030'].get_info(pres['from'], node,
                                                             ifrom=ifrom)

            if isinstance(caps, Iq):
                caps = caps['disco_info']

            if await self._validate_caps(caps, pres['caps']['hash'],
                                               pres['caps']['ver']):
                await self.assign_verstring(pres['from'], pres['caps']['ver'])
        except XMPPError:
            log.debug("Could not retrieve disco#info results for caps for %s", node)

    async def _validate_caps(self, caps, hash, check_verstring):
        # Check Identities
        full_ids = caps.get_identities(dedupe=False)
        deduped_ids = caps.get_identities()
        if len(full_ids) != len(deduped_ids):
            log.debug("Duplicate disco identities found, invalid for caps")
            return False

        # Check Features
        full_features = caps.get_features(dedupe=False)
        deduped_features = caps.get_features()
        if len(full_features) != len(deduped_features):
            log.debug("Duplicate disco features found, invalid for caps")
            return False

        # Check Forms
        form_types = []
        deduped_form_types = set()
        for stanza in caps['substanzas']:
            if not isinstance(stanza, self.xmpp['xep_0004'].stanza.Form):
                log.debug("Non form extension found, ignoring for caps")
                caps.xml.remove(stanza.xml)
                continue
            if 'FORM_TYPE' in stanza.get_fields():
                f_type = tuple(stanza.get_fields()['FORM_TYPE']['value'])
                form_types.append(f_type)
                deduped_form_types.add(f_type)
                if len(form_types) != len(deduped_form_types):
                    log.debug("Duplicated FORM_TYPE values, " + \
                              "invalid for caps")
                    return False

                if len(f_type) > 1:
                    deduped_type = set(f_type)
                    if len(f_type) != len(deduped_type):
                        log.debug("Extra FORM_TYPE data, invalid for caps")
                        return False

                if stanza.get_fields()['FORM_TYPE']['type'] != 'hidden':
                    log.debug("Field FORM_TYPE type not 'hidden', " + \
                              "ignoring form for caps")
                    caps.xml.remove(stanza.xml)
            else:
                log.debug("No FORM_TYPE found, ignoring form for caps")
                caps.xml.remove(stanza.xml)

        verstring = self.generate_verstring(caps, hash)
        if verstring != check_verstring:
            log.debug("Verification strings do not match: %s, %s" % (
                verstring, check_verstring))
            return False

        await self.cache_caps(verstring, caps)
        return True

    def generate_verstring(self, info, hash):
        hash = self.hashes.get(hash, None)
        if hash is None:
            return None

        S = ''

        # Convert None to '' in the identities
        def clean_identity(id):
            return map(lambda i: i or '', id)
        identities = map(clean_identity, info['identities'])

        identities = sorted(('/'.join(i) for i in identities))
        features = sorted(info['features'])

        S += '<'.join(identities) + '<'
        S += '<'.join(features) + '<'

        form_types = {}

        for stanza in info['substanzas']:
            if isinstance(stanza, self.xmpp['xep_0004'].stanza.Form):
                if 'FORM_TYPE' in stanza.get_fields():
                    f_type = stanza['values']['FORM_TYPE']
                    if len(f_type):
                        f_type = f_type[0]
                    if f_type not in form_types:
                        form_types[f_type] = []
                    form_types[f_type].append(stanza)

        sorted_forms = sorted(form_types.keys())
        for f_type in sorted_forms:
            for form in form_types[f_type]:
                S += '%s<' % f_type
                fields = sorted(form.get_fields().keys())
                fields.remove('FORM_TYPE')
                for field in fields:
                    S += '%s<' % field
                    vals = form.get_fields()[field].get_value(convert=False)
                    if vals is None:
                        S += '<'
                    else:
                        if not isinstance(vals, list):
                            vals = [vals]
                        S += '<'.join(sorted(vals)) + '<'

        binary = hash(S.encode('utf8')).digest()
        return base64.b64encode(binary).decode('utf-8')

    async def update_caps(self, jid: OptJidStr = None,
                          node: str | None = None,
                          preserve: bool = False,
                          broadcast: bool = True):
        """Update caps for a local JID based on current data.

        :param jid: JID whose info to update
        :param node: Node to fetch info from
        :param broadcast: Send a presence after updating.
        :param preserve: Send presence only to contacts found in the roster.
        """
        try:
            info = await self.xmpp['xep_0030'].get_info(jid, node, local=True)
            if isinstance(info, Iq):
                info = info['disco_info']
            ver = self.generate_verstring(info, self.hash)
            await self.xmpp['xep_0030'].set_info(
                jid=jid,
                node='%s#%s' % (self.caps_node, ver),
                info=info
            )
            await self.cache_caps(ver, info)
            await self.assign_verstring(jid, ver)

            if broadcast and self.xmpp.sessionstarted and self.broadcast:
                if self.xmpp.is_component or preserve:
                    for contact in self.xmpp.roster[jid]:
                        self.xmpp.roster[jid][contact].send_last_presence()
                else:
                    self.xmpp.roster[jid].send_last_presence()
        except XMPPError:
            return

    def get_verstring(self, jid=None) -> Future:
        """Get the stored verstring for a JID.

        .. versionchanged:: 1.8.0
            This function now returns a Future.
        """
        if jid in ('', None):
            jid = self.xmpp.boundjid.full
        if isinstance(jid, JID):
            jid = jid.full
        return self.api['get_verstring'](jid)

    def assign_verstring(self, jid=None, verstring=None) -> Future:
        """Assign a verification string to a jid.

        .. versionchanged:: 1.8.0
            This function now returns a Future.
        """
        if jid in (None, ''):
            jid = self.xmpp.boundjid.full
        if isinstance(jid, JID):
            jid = jid.full
        return self.api['assign_verstring'](jid, args={
            'verstring': verstring
        })

    def cache_caps(self, verstring=None, info=None) -> Future:
        """Add caps to the cache.

        .. versionchanged:: 1.8.0
            This function now returns a Future.
        """
        data = {'verstring': verstring, 'info': info}
        return self.api['cache_caps'](args=data)

    async def get_caps(self, jid=None, verstring=None):
        """Get caps for a JID.

        .. versionchanged:: 1.8.0
            This function is now a coroutine.
        """
        if verstring is None:
            if jid is not None:
                verstring = await self.get_verstring(jid)
            else:
                return None
        if isinstance(jid, JID):
            jid = jid.full
        data = {'verstring': verstring}
        return await self.api['get_caps'](jid, args=data)



"""
Stanza interfaces for XEP-0449: Stickers

Basic usage:

>>> register()  # automatically done if you use this plugin
>>> msg = Message()
>>> msg["sticker"]["pack"] = "some-pack-id"
>>> msg.pretty_print()
<message xmlns="jabber:client">
  <sticker xmlns="urn:xmpp:stickers:0" pack="some-pack-id" />
</message>

You are supposed to also add a Stateless File Sharing payload if you do that.
"""

from slixmpp import JID, InvalidJID, Message
from slixmpp.types import JidStr
from slixmpp.xmlstream import ElementBase, register_stanza_plugin

NAMESPACE = "urn:xmpp:stickers:0"


class Sticker(ElementBase):
    name = "sticker"
    plugin_attrib = "sticker"
    namespace = NAMESPACE
    interfaces = {"pack", "jid", "node"}

    def get_jid(self) -> JID | None:
        try:
            return JID(self._get_attr("jid"))
        except InvalidJID:
            return None

    def set_jid(self, jid: JidStr) -> None:
        self._set_attr("jid", str(jid))


def register() -> None:
    register_stanza_plugin(Message, Sticker)



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2020 Mathieu Pasquet
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from typing import Iterable

from slixmpp import JID
from slixmpp.plugins import BasePlugin
from slixmpp.stanza import Message
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.xmlstream.matcher import StanzaPath
from slixmpp.xmlstream.handler import Callback

from slixmpp.plugins.xep_0444 import stanza


class XEP_0444(BasePlugin):
    """
    XEP-0444: Message Reactions.

    If the python-emoji library is present, setting emojis as reactions will
    be checked against known emoji, and trying to set non-emoji characters
    as reactions will raise a ``ValueError``. This behavior can be disabled
    by passing ``all_chars=True`` to the ``Reaction.set_value()`` call.
    """
    name = 'xep_0444'
    description = 'XEP-0444: Message Reactions'
    dependencies = {'xep_0030', 'xep_0334'}
    stanza = stanza
    namespace = stanza.NS

    def plugin_init(self):
        self.xmpp.register_handler(Callback(
            'Reaction received',
            StanzaPath("message/reactions"),
            self._handle_reactions,
        ))
        register_stanza_plugin(Message, stanza.Reactions)
        register_stanza_plugin(stanza.Reactions, stanza.Reaction, iterable=True)

    def session_bind(self, event):
        self.xmpp['xep_0030'].add_feature(stanza.NS)

    def plugin_end(self):
        self.xmpp.remove_handler('Reaction received')
        self.xmpp['xep_0030'].del_feature(feature=stanza.NS)

    def _handle_reactions(self, message: Message):
        self.xmpp.event('reactions', message)

    def send_reactions(self, to: JID, to_id: str, reactions: Iterable[str], *, store=True):
        """
        Send reactions related to a message.

        :param to: JID to which the reactions will be sent.
        :param to_id: ID of the message being reacted to.
        :param reactions: Iterable object containing all reactions to set.
        :param store: If the store hint should be set on the message.
        """
        msg = self.xmpp.make_message(mto=to)
        self.set_reactions(msg, to_id, reactions)
        if store:
            msg.enable('store')
        msg.send()

    @staticmethod
    def set_reactions(message: Message, to_id: str, reactions: Iterable[str]):
        """
        Add reactions to a Message object.

        :param message: Message to enrich with reactions.
        :param to_id: ID of the message the reactions are for.
        :param reactions: Iterable object containing all reactions to set.
        """
        message['reactions']['id'] = to_id
        for reaction in reactions:
            reaction_stanza = stanza.Reaction()
            reaction_stanza['value'] = reaction
            message['reactions'].append(reaction_stanza)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permissio
from slixmpp.jid import JID
from slixmpp.xmlstream import ElementBase, register_stanza_plugin


class Offline(ElementBase):
    name = 'offline'
    namespace = 'http://jabber.org/protocol/offline'
    plugin_attrib = 'offline'
    interfaces = {'fetch', 'purge', 'results'}
    bool_interfaces = interfaces

    def setup(self, xml=None):
        ElementBase.setup(self, xml)
        self._results = []

    # The results interface is meant only as an easy
    # way to access the set of collected message responses
    # from the query.

    def get_results(self):
        return self._results

    def set_results(self, values):
        self._results = values

    def del_results(self):
        self._results = []


class Item(ElementBase):
    name = 'item'
    namespace = 'http://jabber.org/protocol/offline'
    plugin_attrib = 'item'
    interfaces = {'action', 'node', 'jid'}

    actions = {'view', 'remove'}

    def get_jid(self):
        return JID(self._get_attr('jid'))

    def set_jid(self, value):
        self._set_attr('jid', str(value))


register_stanza_plugin(Offline, Item, iterable=True)



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permissio
import logging

from slixmpp.stanza import Message
from slixmpp.jid import JID
from slixmpp.xmlstream.handler import Callback
from slixmpp.xmlstream.matcher import StanzaPath
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.plugins import BasePlugin
from slixmpp.plugins.xep_0308 import stanza, Replace


log = logging.getLogger(__name__)


class XEP_0308(BasePlugin):

    """
    XEP-0308 Last Message Correction
    """

    name = 'xep_0308'
    description = 'XEP-0308: Last Message Correction'
    dependencies = {'xep_0030'}
    stanza = stanza

    def plugin_init(self):
        self.xmpp.register_handler(
            Callback('Message Correction',
                     StanzaPath('message/replace'),
                     self._handle_correction))

        register_stanza_plugin(Message, Replace)

        self.xmpp.use_message_ids = True

    def plugin_end(self):
        self.xmpp.remove_handler('Message Correction')
        self.xmpp.plugin['xep_0030'].del_feature(feature=Replace.namespace)

    def session_bind(self, jid):
        self.xmpp.plugin['xep_0030'].add_feature(Replace.namespace)

    def is_correction(self, msg: Message) -> bool:
        """
        Check if a message contains a correction.

        :param msg: Message to check.
        :returns: True if the message contains a correction.
        """
        return msg.xml.find('{%s}replace' % Replace.namespace) is not None

    def _handle_correction(self, msg: Message):
        self.xmpp.event('message_correction', msg)

    def build_correction(self, id_to_replace: str, mto: JID,
                         mfrom: JID | None = None, mtype: str = 'chat',
                         mbody: str = '') -> Message:
        """
        Build a corrected message.

        :param id_to_replace: The id of the original message.
        :param mto: Recipient of the message, must be the same as the original
                    message.
        :param mfrom: Sender of the message, must be the same as the original
                      message.
        :param mtype: Type of the message, must be the send as the original
                      message.
        :param mbody: The corrected message body.
        """
        msg = self.xmpp.make_message(
            mto=mto,
            mfrom=mfrom,
            mbody=mbody,
            mtype=mtype
        )
        msg['replace']['id'] = id_to_replace
        return msg

    def correct_message(self, msg: Message, body: str) -> Message:
        """
        Send a correction to an existing message.

        :param msg: The message that must be replaced.
        :param body: The body to set in the correcting message.
        :returns: The message that was sent.
        """
        to_replace = msg['id']
        mto = msg['to']
        mfrom = msg['from']
        mtype = msg['type']
        if not to_replace:
            raise ValueError('No available ID for replacing the message')
        if not mto:
            raise ValueError('No available recipient JID')

        new = self.build_correction(
            id_to_replace=to_replace,
            mto=mto,
            mfrom=mfrom,
            mtype=mtype,
            mbody=body,
        )
        new.send()
        return new




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2011  Nathanael C. Fritz
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.xmlstream import ElementBase, ET


class Session(ElementBase):

    """
    """

    name = 'session'
    namespace = 'urn:ietf:params:xml:ns:xmpp-session'
    interfaces = {'optional'}
    plugin_attrib = 'session'

    def get_optional(self):
        return self.xml.find('{%s}optional' % self.namespace) is not None

    def set_optional(self, value):
        if value:
            optional = ET.Element('{%s}optional' % self.namespace)
            self.xml.append(optional)
        else:
            self.del_optional()

    def del_optional(self):
        optional = self.xml.find('{%s}optional' % self.namespace)
        self.xml.remove(optional)



# slixmpp.xmlstream.matcher.base
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Part of Slixmpp: The Slick XMPP Library
# :copyright: (c) 2011 Nathanael C. Fritz
# :license: MIT, see LICENSE for more details

from typing import Any
from slixmpp.xmlstream.stanzabase import StanzaBase


class MatcherBase(object):

    """
    Base class for stanza matchers. Stanza matchers are used to pick
    stanzas out of the XML stream and pass them to the appropriate
    stream handlers.

    :param criteria: Object to compare some aspect of a stanza against.
    """

    def __init__(self, criteria: Any):
        self._criteria = criteria

    def match(self, xml: StanzaBase) -> bool:
        """Check if a stanza matches the stored criteria.

        Meant to be overridden.
        """
        return False



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp import Iq
from slixmpp.xmlstream import ElementBase, ET, register_stanza_plugin
from collections.abc import Sequence


class Certs(ElementBase):
    """
    List of certs returned by the server.
    This element is only useful for accessing the certs inside.

    .. code-block:: xml

          <items xmlns='urn:xmpp:saslcert:1'>
            <item>
              <name>Mobile Client</name>
              <x509cert>
                ...
              </x509cert>
              <users>
                <resource>Phone</resource>
              </users>
            </item>
            <item>
              <name>Laptop</name>
              <x509cert>
                ...
              </x509cert>
            </item>
          </items>
        </iq>
    """
    name = 'items'
    namespace = 'urn:xmpp:saslcert:1'
    plugin_attrib = 'sasl_certs'
    interfaces = set()


class CertItem(ElementBase):
    """
    A single cert item.

    Contains the X.509 base64 DER representation, its name, as well
    as a list of users currently logged in with it.

    .. code-block:: xml

        <item xmlns='urn:xmpp:saslcert:1'>
          <name>Mobile Client</name>
          <x509cert>
            ...
          </x509cert>
          <users>
            <resource>Phone</resource>
          </users>
        </item>

    """
    name = 'item'
    namespace = 'urn:xmpp:saslcert:1'
    plugin_attrib = 'item'
    plugin_multi_attrib = 'items'
    interfaces = {'name', 'x509cert', 'users'}
    sub_interfaces = {'name', 'x509cert'}

    def get_users(self) -> set[str]:
        """Return the resources currently using this cert."""
        resources = self.xml.findall('{%s}users/{%s}resource' % (
            self.namespace, self.namespace))
        return {res.text for res in resources if res is not None and res.text}

    def set_users(self, values: Sequence[str]) -> None:
        users = self.xml.find('{%s}users' % self.namespace)
        if users is not None:
            self.xml.remove(users)
        if not values:
            return
        users = ET.Element('{%s}users' % self.namespace)
        self.xml.append(users)
        for resource in values:
            res = ET.Element('{%s}resource' % self.namespace)
            res.text = resource
            users.append(res)

    def del_users(self) -> None:
        users = self.xml.find('{%s}users' % self.namespace)
        if users is not None:
            self.xml.remove(users)


class AppendCert(ElementBase):
    """
    Element used for adding a cert.

    .. code-block:: xml

        <append xmlns='urn:xmpp:saslcert:1'>
            <name>Simple Bot</name>
            <no-cert-management/>
            <x509cert>
              ...
            </x509cert>
        </append>
    """
    name = 'append'
    namespace = 'urn:xmpp:saslcert:1'
    plugin_attrib = 'sasl_cert_append'
    interfaces = {'name', 'x509cert', 'cert_management'}
    sub_interfaces = {'name', 'x509cert'}

    def get_cert_management(self) -> bool:
        manage = self.xml.find('{%s}no-cert-management' % self.namespace)
        return manage is None

    def set_cert_management(self, value: bool) -> None:
        self.del_cert_management()
        if not value:
            manage = ET.Element('{%s}no-cert-management' % self.namespace)
            self.xml.append(manage)

    def del_cert_management(self) -> None:
        manage = self.xml.find('{%s}no-cert-management' % self.namespace)
        if manage is not None:
            self.xml.remove(manage)


class DisableCert(ElementBase):
    """
    Element used for disabling a cert.

    .. code-block:: xml

        <disable xmlns='urn:xmpp:saslcert:1'>
            <name>Mobile Client</name>
        </disable>
    """
    name = 'disable'
    namespace = 'urn:xmpp:saslcert:1'
    plugin_attrib = 'sasl_cert_disable'
    interfaces = {'name'}
    sub_interfaces = interfaces


class RevokeCert(ElementBase):
    """
    Element used for revoking a cert.

    .. code-block:: xml

        <revoke xmlns='urn:xmpp:saslcert:1'>
            <name>Mobile Client</name>
        </revoke>
    """
    name = 'revoke'
    namespace = 'urn:xmpp:saslcert:1'
    plugin_attrib = 'sasl_cert_revoke'
    interfaces = {'name'}
    sub_interfaces = interfaces


def register_plugins():
    register_stanza_plugin(Certs, CertItem, iterable=True)
    register_stanza_plugin(Iq, Certs)
    register_stanza_plugin(Iq, AppendCert)
    register_stanza_plugin(Iq, DisableCert)
    register_stanza_plugin(Iq, RevokeCert)



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2020 Mathieu Pasquet <mathieui@mathieui.net>
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins.base import register_plugin
from slixmpp.plugins.xep_0422.stanza import ApplyTo, External
from slixmpp.plugins.xep_0422.fastening import XEP_0422

register_plugin(XEP_0422)

__all__ = ['ApplyTo', 'External', 'XEP_0422']




# slixmpp.util.sasl
# ~~~~~~~~~~~~~~~~~~~
# This module was originally based on Dave Cridland's Suelta library.
# Part of Slixmpp: The Slick XMPP Library
# :copyright: (c) 2004-2013 David Alan Cridland
# :copyright: (c) 2013 Nathanael C. Fritz, Lance J.T. Stout
# :license: MIT, see LICENSE for more details
from slixmpp.util.sasl.client import *
from slixmpp.util.sasl.mechanisms import *




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2011 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.types import OptJidStr, OptJid
import asyncio
import functools
import logging
import time
import typing
from inspect import iscoroutinefunction

from slixmpp import Iq
from slixmpp.exceptions import XMPPError
from slixmpp.xmlstream.handler import Callback
from slixmpp.xmlstream.matcher import StanzaPath
from slixmpp.xmlstream import register_stanza_plugin, JID
from slixmpp.plugins import BasePlugin
from slixmpp.plugins.xep_0050 import stanza
from slixmpp.plugins.xep_0050 import Command
from slixmpp.plugins.xep_0004 import Form


log = logging.getLogger(__name__)


SessionDict = dict[str, typing.Any]
HandlerType = typing.Callable[[Iq, SessionDict], typing.Awaitable[SessionDict] | SessionDict]
TimeoutHandlerType = typing.Callable[[SessionDict], typing.Awaitable[None]]
CommandType = tuple[str, HandlerType | None, TimeoutHandlerType | None, float]


class XEP_0050(BasePlugin):

    """
    XEP-0050: Ad-Hoc Commands

    XMPP's Adhoc Commands provides a generic workflow mechanism for
    interacting with applications. The result is similar to menu selections
    and multi-step dialogs in normal desktop applications. Clients do not
    need to know in advance what commands are provided by any particular
    application or agent. While adhoc commands provide similar functionality
    to Jabber-RPC, adhoc commands are used primarily for human interaction.

    Also see <http://xmpp.org/extensions/xep-0050.html>

    Events:
        - command_execute: Received a command with action="execute"
        - command_next: Received a command with action="next"
        - command_complete: Received a command with action="complete"
        - command_cancel: Received a command with action="cancel"

    Attributes:
        - ``commands``: A dictionary mapping JID/node pairs to command
            names and handlers. Only used with the default API handlers.
        - ``sessions``: A dictionary or equivalent backend mapping
            session IDs to dictionaries containing data
            relevant to a command's session.

    Slixmpp API:
        A basic default static implementation is provided for those using the
        ``self.commands`` attribute.

        - set_command: Set handler for a ``jid/node/ifrom`` combo
        - get_command: Get handler for a ``jid/node/ifrom`` combo

    """

    name = 'xep_0050'
    description = 'XEP-0050: Ad-Hoc Commands'
    dependencies = {'xep_0030', 'xep_0004'}
    stanza = stanza
    default_config = {
        'session_db': None
    }
    commands: dict[tuple[str , str | None], CommandType]
    _timeout_tasks: dict[str, asyncio.Task]

    def plugin_init(self):
        """Start the XEP-0050 plugin."""
        self.sessions = self.session_db
        if self.sessions is None:
            self.sessions = {}

        self.commands = {}
        self._timeout_tasks = {}

        self.xmpp.register_handler(
            Callback("Ad-Hoc Execute",
                     StanzaPath('iq@type=set/command'),
                     self._handle_command))

        register_stanza_plugin(Iq, Command)
        register_stanza_plugin(Command, Form, iterable=True)

        self.xmpp.add_event_handler('command', self._handle_command_all)
        self.api.register(self._set_command, "set_command", default=True)
        self.api.register(self._get_command, "get_command", default=True)

    def plugin_end(self):
        self.xmpp.del_event_handler('command', self._handle_command_all)
        self.xmpp.remove_handler('Ad-Hoc Execute')
        self.xmpp['xep_0030'].del_feature(feature=Command.namespace)
        self.xmpp['xep_0030'].set_items(node=Command.namespace, items=tuple())

    def session_bind(self, jid):
        self.xmpp['xep_0030'].add_feature(Command.namespace)
        self.xmpp['xep_0030'].set_items(node=Command.namespace, items=tuple())

    def set_backend(self, db):
        """
        Replace the default session storage dictionary with
        a generic, external data storage mechanism.

        The replacement backend must be able to interact through
        the same syntax and interfaces as a normal dictionary.

        :param db: The new session storage mechanism.
        """
        self.sessions = db

    def prep_handlers(self, handlers, **kwargs):
        """
        Prepare a list of functions for use by the backend service.

        Intended to be replaced by the backend service as needed.

        :param handlers: A list of function pointers
        :param kwargs: Any additional parameters required by the backend.
        """
        pass

    # Default API handlers

    async def _set_command(self,
        jid: OptJid = None,
        node: str | None = None,
        ifrom: OptJid = None,
        args: CommandType | None = None,
    ) -> None:
        if args is None:
            raise RuntimeError
        jid = self.__default_jid(jid)
        self.commands[(jid.full, node)] = args

    async def _get_command(self,
        jid: OptJidStr = None,
        node: str | None = None,
        ifrom: OptJid = None,
        args: None = None,
    ) -> CommandType | None:
        jid = self.__default_jid(jid)
        return self.commands.get((jid.full, node))

    def __default_jid(self, jid: OptJidStr) -> JID:
        if jid is None:
            jid = self.xmpp.boundjid
        elif not isinstance(jid, JID):
            jid = JID(jid)
        return jid

    # =================================================================
    # Server side (command provider) API

    def add_command(self,
                    jid:OptJidStr = None,
                    node:str | None = None,
                    name='',
                    handler: HandlerType | None=None,
                    *,
                    timeout: float = 0,
                    timeout_handler: TimeoutHandlerType | None = None):
        """
        Make a new command available to external entities.

        Access control may be implemented in the provided handler.

        Command workflow is done across a sequence of command handlers. The
        first handler is given the initial Iq stanza of the request in order
        to support access control. Subsequent handlers are given only the
        payload items of the command. All handlers will receive the command's
        session data.

        :param jid: The JID that will expose the command.
        :param node: The node associated with the command.
        :param name: A human readable name for the command.
        :param handler: A function that will generate the response to the
                        initial command request, as well as enforcing any
                        access control policies.
        :param timeout: A timeout, in seconds, after which a session for
            this command will be cancelled. 0 by default, which means no timeout.
        :param timeout_handler: Optionally, a custom handler that will be called
            on timeout, with the session dict, before the "cancel" IQ is sent
            to the XMPP client.
        """
        jid = self.__default_jid(jid)
        item_jid = jid.full

        self.xmpp['xep_0030'].add_identity(category='automation',
                                           itype='command-list',
                                           name='Ad-Hoc commands',
                                           node=Command.namespace,
                                           jid=jid)
        self.xmpp['xep_0030'].add_item(jid=item_jid,
                                       name=name,
                                       node=Command.namespace,
                                       subnode=node,
                                       ijid=jid)
        self.xmpp['xep_0030'].add_identity(category='automation',
                                           itype='command-node',
                                           name=name,
                                           node=node,
                                           jid=jid)
        self.xmpp['xep_0030'].add_feature(Command.namespace, None, jid)

        self.xmpp.loop.create_task(
            self.__add_command(item_jid, node, None, (name, handler, timeout_handler, timeout))
        )

    async def __add_command(self, *args) -> None:
        await self.api["set_command"](*args)

    def new_session(self):
        """Return a new session ID."""
        return str(time.time()) + '-' + self.xmpp.new_id()

    def _handle_command(self, iq):
        """Raise command events based on the command action."""
        self.xmpp.event('command', iq)
        self.xmpp.event('command_%s' % iq['command']['action'], iq)

    async def _handle_command_all(self, iq: Iq) -> None:
        action = iq['command']['action']
        sessionid = iq['command']['sessionid']
        session = self.sessions.get(sessionid)

        if session is None:
            return await self._handle_command_start(iq)

        if action in ('next', 'execute'):
            return await self._handle_command_next(iq)
        if action == 'prev':
            return await self._handle_command_prev(iq)
        if action == 'complete':
            return await self._handle_command_complete(iq)
        if action == 'cancel':
            return await self._handle_command_cancel(iq)
        return None

    async def _handle_command_start(self, iq: Iq) -> None:
        """
        Process an initial request to execute a command.

        :param iq: The command execution request.
        """
        sessionid = self.new_session()
        node = iq['command']['node']
        key = (iq['to'].full, node)

        name, handler, _, _ = await self.api["get_command"](*key, iq.get_from()) or ('Not found', None, None, 0)
        if not handler:
            log.debug('Command not found: %s', key)
            raise XMPPError('item-not-found')

        payload = []
        for stanza_ in iq['command']['substanzas']:
            payload.append(stanza_)

        if len(payload) == 1:
            payload = payload[0]

        interfaces = {item.plugin_attrib for item in payload}
        payload_classes = {item.__class__ for item in payload}

        initial_session = {'id': sessionid,
                           'from': iq['from'],
                           'to': iq['to'],
                           'node': node,
                           'payload': payload,
                           'interfaces': interfaces,
                           'payload_classes': payload_classes,
                           'notes': None,
                           'has_next': False,
                           'allow_complete': False,
                           'allow_prev': False,
                           'past': [],
                           'next': None,
                           'prev': None,
                           'cancel': None}

        await self._reset_timeout_task(initial_session, iq)
        session = await _await_if_needed(handler, iq, initial_session)

        await self._process_command_response(iq, session)

    async def _handle_command_next(self, iq):
        """
        Process a request for the next step in the workflow
        for a command with multiple steps.

        :param iq: The command continuation request.
        """
        sessionid = iq['command']['sessionid']
        session = self.sessions.get(sessionid)
        await self._reset_timeout_task(session, iq)

        if session:
            handler = session['next']
            interfaces = session['interfaces']
            results = []
            for stanza in iq['command']['substanzas']:
                if stanza.plugin_attrib in interfaces:
                    results.append(stanza)
            if len(results) == 1:
                results = results[0]

            session = await _await_if_needed(handler, results, session)

            await self._process_command_response(iq, session)
        else:
            raise XMPPError('item-not-found')

    async def _handle_command_prev(self, iq):
        """
        Process a request for the prev step in the workflow
        for a command with multiple steps.

        :param iq: The command continuation request.
        """
        sessionid = iq['command']['sessionid']
        session = self.sessions.get(sessionid)
        await self._reset_timeout_task(session, iq)

        if session:
            handler = session['prev']
            interfaces = session['interfaces']
            results = []
            for stanza in iq['command']['substanzas']:
                if stanza.plugin_attrib in interfaces:
                    results.append(stanza)
            if len(results) == 1:
                results = results[0]

            session = await _await_if_needed(handler, results, session)

            await self._process_command_response(iq, session)
        else:
            raise XMPPError('item-not-found')

    async def _process_command_response(self, iq: Iq, session: dict) -> None:
        """
        Generate a command reply stanza based on the
        provided session data.

        :param iq: The command request stanza.
        :param session: A dictionary of relevant session data.
        """
        sessionid = session['id']
        await self._reset_timeout_task(session, iq)

        payload = session['payload']
        if payload is None:
            payload = []
        if not isinstance(payload, list):
            payload = [payload]

        interfaces = session.get('interfaces', set())
        payload_classes = session.get('payload_classes', set())

        interfaces.update({item.plugin_attrib for item in payload})
        payload_classes.update({item.__class__ for item in payload})

        session['interfaces'] = interfaces
        session['payload_classes'] = payload_classes

        self.sessions[sessionid] = session

        for item in payload:
            register_stanza_plugin(Command, item.__class__, iterable=True)

        iq = iq.reply()
        iq['command']['node'] = session['node']
        iq['command']['sessionid'] = session['id']

        if session['next'] is None:
            iq['command']['actions'] = []
            iq['command']['status'] = 'completed'
        elif session['has_next']:
            actions = ['next']
            if session['allow_complete']:
                actions.append('complete')
            if session['allow_prev']:
                actions.append('prev')
            iq['command']['actions'] = actions
            iq['command']['status'] = 'executing'
        else:
            actions = ['complete']
            if session['allow_prev']:
                actions.append('prev')
            iq['command']['actions'] = actions
            iq['command']['status'] = 'executing'

        iq['command']['notes'] = session['notes']

        for item in payload:
            iq['command'].append(item)

        iq.send()

    async def _handle_command_cancel(self, iq):
        """
        Process a request to cancel a command's execution.

        :param iq: The command cancellation request.
        """
        node = iq['command']['node']
        sessionid = iq['command']['sessionid']
        self._cancel_timeout_task(sessionid)

        session = self.sessions.get(sessionid)

        if session:
            handler = session['cancel']
            if handler:
                await _await_if_needed(handler, iq, session)
            del self.sessions[sessionid]
            iq = iq.reply()
            iq['command']['node'] = node
            iq['command']['sessionid'] = sessionid
            iq['command']['status'] = 'canceled'
            iq['command']['notes'] = session['notes']
            iq.send()
        else:
            raise XMPPError('item-not-found')


    async def _handle_command_complete(self, iq):
        """
        Process a request to finish the execution of command
        and terminate the workflow.

        All data related to the command session will be removed.

        Arguments:
        :param iq: The command completion request.
        """
        node = iq['command']['node']
        sessionid = iq['command']['sessionid']
        session = self.sessions.get(sessionid)
        self._cancel_timeout_task(sessionid)

        if session:
            handler = session['next']
            interfaces = session['interfaces']
            results = []
            for stanza in iq['command']['substanzas']:
                if stanza.plugin_attrib in interfaces:
                    results.append(stanza)
            if len(results) == 1:
                results = results[0]

            if handler:
                await _await_if_needed(handler, results, session)

            del self.sessions[sessionid]

            payload = session['payload']
            if payload is None:
                payload = []
            if not isinstance(payload, list):
                payload = [payload]

            for item in payload:
                register_stanza_plugin(Command, item.__class__, iterable=True)

            iq = iq.reply()

            iq['command']['node'] = node
            iq['command']['sessionid'] = sessionid
            iq['command']['actions'] = []
            iq['command']['status'] = 'completed'
            iq['command']['notes'] = session['notes']

            for item in payload:
                iq['command'].append(item)

            iq.send()
        else:
            raise XMPPError('item-not-found')

    async def _reset_timeout_task(self, session: SessionDict | None, iq: Iq) -> None:
        if not session:
            return

        key = (iq['to'].full, iq['command']['node'])
        _, _, handler, timeout = await self.api["get_command"](*key, iq.get_from()) or ("Not found", None, None, 0)

        self._cancel_timeout_task(session["id"])

        if not timeout:
            return

        async def timer() -> None:
            await asyncio.sleep(timeout)
            if handler is not None:
                await handler(session)
            session['jid'] = iq.get_from()
            self.cancel_command(session)
            self._cancel_timeout_task(session["id"])

        self._timeout_tasks[session["id"]] = asyncio.create_task(timer())

    def _cancel_timeout_task(self, sessionid: str) -> None:
        existing = self._timeout_tasks.pop(sessionid, None)
        if existing is not None and not existing.done():
            existing.cancel()

    # =================================================================
    # Client side (command user) API

    def get_commands(self, jid, **kwargs):
        """
        Return a list of commands provided by a given JID.

        :param jid: The JID to query for commands.
        :param local: If true, then the query is for a JID/node
                      combination handled by this Slixmpp instance and
                      no stanzas need to be sent.
                      Otherwise, a disco stanza must be sent to the
                      remote JID to retrieve the items.
        :param iterator: If True, return a result set iterator using
                         the XEP-0059 plugin, if the plugin is loaded.
                         Otherwise the parameter is ignored.
        """
        return self.xmpp['xep_0030'].get_items(jid=jid,
                                               node=Command.namespace,
                                               **kwargs)

    def send_command(self, jid, node, ifrom=None, action='execute',
                     payload=None, sessionid=None, flow=False, **kwargs):
        """
        Create and send a command stanza, without using the provided
        workflow management APIs.

        :param jid: The JID to send the command request or result.
        :param node: The node for the command.
        :param ifrom: Specify the sender's JID.
        :param action: May be one of: execute, cancel, complete,
                         or cancel.
        :param payload: Either a list of payload items, or a single
                        payload item such as a data form.
        :param sessionid: The current session's ID value.
        :param flow: If True, process the Iq result using the
                     command workflow methods contained in the
                     session instead of returning the response
                     stanza itself. Defaults to False.
        """
        iq = self.xmpp.Iq()
        iq['type'] = 'set'
        iq['to'] = jid
        iq['from'] = ifrom
        iq['command']['node'] = node
        iq['command']['action'] = action
        if sessionid is not None:
            iq['command']['sessionid'] = sessionid
        if payload is not None:
            if not isinstance(payload, list):
                payload = [payload]
            for item in payload:
                iq['command'].append(item)
        if not flow:
            return iq.send(**kwargs)
        else:
            iq.send(callback=self._handle_command_result)

    def start_command(self, jid, node, session, ifrom=None):
        """
        Initiate executing a command provided by a remote agent.

        The provided session dictionary should contain:

        :param next: A handler for processing the command result.
        :param error: A handler for processing any error stanzas
                      generated by the request.

        :param jid: The JID to send the command request.
        :param node: The node for the desired command.
        :param session: A dictionary of relevant session data.
        """
        session['jid'] = jid
        session['node'] = node
        session['timestamp'] = time.time()
        if 'payload' not in session:
            session['payload'] = None

        iq = self.xmpp.Iq()
        iq['type'] = 'set'
        iq['to'] = jid
        iq['from'] = ifrom
        session['from'] = ifrom
        iq['command']['node'] = node
        iq['command']['action'] = 'execute'
        if session['payload'] is not None:
            payload = session['payload']
            if not isinstance(payload, list):
                payload = list(payload)
            for stanza in payload:
                iq['command'].append(stanza)
        sessionid = 'client:pending_' + iq['id']
        session['id'] = sessionid
        self.sessions[sessionid] = session
        iq.send(callback=self._handle_command_result)

    def continue_command(self, session, direction='next'):
        """
        Execute the next action of the command.

        :param session: All stored data relevant to the current
                        command session.
        """
        sessionid = 'client:' + session['id']
        self.sessions[sessionid] = session

        self.send_command(session['jid'],
                          session['node'],
                          ifrom=session.get('from', None),
                          action=direction,
                          payload=session.get('payload', None),
                          sessionid=session['id'],
                          flow=True)

    def cancel_command(self, session):
        """
        Cancel the execution of a command.

        :param session: All stored data relevant to the current
                        command session.
        """
        sessionid = 'client:' + session['id']
        self.sessions[sessionid] = session

        self.send_command(session['jid'],
                          session['node'],
                          ifrom=session.get('from', None),
                          action='cancel',
                          payload=session.get('payload', None),
                          sessionid=session['id'],
                          flow=True)

    def complete_command(self, session):
        """
        Finish the execution of a command workflow.

        :param session: All stored data relevant to the current
                        command session.
        """
        sessionid = 'client:' + session['id']
        self.sessions[sessionid] = session

        self.send_command(session['jid'],
                          session['node'],
                          ifrom=session.get('from', None),
                          action='complete',
                          payload=session.get('payload', None),
                          sessionid=session['id'],
                          flow=True)

    def terminate_command(self, session):
        """
        Delete a command's session after a command has completed
        or an error has occurred.

        :param session: All stored data relevant to the current
                        command session.
        """
        sessionid = 'client:' + session['id']
        try:
            del self.sessions[sessionid]
        except Exception as e:
            log.error("Error deleting adhoc command session: %s" % e)

    def _handle_command_result(self, iq):
        """
        Process the results of a command request.

        Will execute the 'next' handler stored in the session
        data, or the 'error' handler depending on the Iq's type.

        :param iq: The command response.
        """
        sessionid = 'client:' + iq['command']['sessionid']
        pending = False

        if sessionid not in self.sessions:
            pending = True
            pendingid = 'client:pending_' + iq['id']
            if pendingid not in self.sessions:
                return
            sessionid = pendingid

        session = self.sessions[sessionid]
        sessionid = 'client:' + iq['command']['sessionid']
        session['id'] = iq['command']['sessionid']

        self.sessions[sessionid] = session

        if pending:
            del self.sessions[pendingid]

        handler_type = 'next'
        if iq['type'] == 'error':
            handler_type = 'error'
        handler = session.get(handler_type, None)
        if handler:
            handler(iq, session)
        elif iq['type'] == 'error':
            self.terminate_command(session)

        if iq['command']['status'] == 'completed':
            self.terminate_command(session)


def _iscoroutine_or_partial_coroutine(handler):
    return iscoroutinefunction(handler) \
        or isinstance(handler, functools.partial) \
        and iscoroutinefunction(handler.func)


async def _await_if_needed(handler, *args):
    if handler is None:
        raise XMPPError("bad-request", text="The command is completed")
    if _iscoroutine_or_partial_coroutine(handler):
        log.debug("%s is async", handler)
        return await handler(*args)
    else:
        log.debug("%s is sync", handler)
        return handler(*args)



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2020 Mathieu Pasquet <mathieui@mathieui.net>
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins.base import register_plugin
from slixmpp.plugins.xep_0405.stanza import (
    ClientJoin, ClientLeave, Annotate, Channel,
)
from slixmpp.plugins.xep_0405.mix_pam import XEP_0405

register_plugin(XEP_0405)

__all__ = ['ClientJoin', 'ClientLeave', 'Annotate', 'Channel',
           'XEP_0405']




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permissio
from slixmpp.xmlstream import ElementBase


class Replace(ElementBase):
    """
    Replace element, containing the id of the message to replace.

    .. code-block:: xml

        <replace id='bad1' xmlns='urn:xmpp:message-correct:0'/>
    """
    name = 'replace'
    namespace = 'urn:xmpp:message-correct:0'
    plugin_attrib = 'replace'
    interfaces = {'id'}



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2016 Emmanuel Gil Peyrot <linkmauve@linkmauve.fr>
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins.base import register_plugin

from slixmpp.plugins.xep_0380.stanza import Encryption
from slixmpp.plugins.xep_0380.eme import XEP_0380


register_plugin(XEP_0380)

__all__ = ['Encryption', ' XEP_0380']




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import logging

from slixmpp import ClientXMPP
from slixmpp.stanza import StreamFeatures
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.plugins.base import BasePlugin
from slixmpp.plugins.xep_0352 import stanza, ClientStateIndication


log = logging.getLogger(__name__)


class XEP_0352(BasePlugin):

    """
    XEP-0352: Client State Indication
    """

    name = 'xep_0352'
    description = 'XEP-0352: Client State Indication'
    dependencies = set()
    stanza = stanza
    default_config = {
        "order": 12000,
    }

    def plugin_init(self):
        """Start the XEP-0352 plugin."""

        self.enabled = False

        register_stanza_plugin(StreamFeatures, ClientStateIndication)
        self.xmpp.register_stanza(stanza.Active)
        self.xmpp.register_stanza(stanza.Inactive)

        if isinstance(self.xmpp, ClientXMPP):
            self.xmpp.register_feature('csi',
                    self._handle_csi_feature,
                    restart=False,
                    order=self.order)


    def plugin_end(self):
        if self.xmpp.is_component:
            return

        if isinstance(self.xmpp, ClientXMPP):
            self.xmpp.unregister_feature('csi', self.order)
        self.xmpp.remove_stanza(stanza.Active)
        self.xmpp.remove_stanza(stanza.Inactive)

    def send_active(self):
        """Send an 'active' state"""
        if self.enabled:
            self.xmpp.send_raw(str(stanza.Active(self.xmpp)))

    def send_inactive(self):
        """Send an 'active' state"""
        if self.enabled:
            self.xmpp.send_raw(str(stanza.Inactive(self.xmpp)))

    def _handle_csi_feature(self, features):
        """
        Enable CSI
        """
        if 'csi' in self.xmpp.features:
            log.debug('CSI already enabled')
            return False
        self.enabled = True
        self.xmpp.event('csi_enabled', features)
        return False



# Slixmpp: The Slick XMPP Library
# Implementation of HTTP over XMPP transport
# http://xmpp.org/extensions/xep-0332.html
# Copyright (C) 2015 Riptide IO, sangeeth@riptideio.com
# This file is part of slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins.base import register_plugin

from slixmpp.plugins.xep_0332 import stanza
from slixmpp.plugins.xep_0332.http import XEP_0332


register_plugin(XEP_0332)

__all__ = ['XEP_0332', 'stanza']




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2011  Nathanael C. Fritz
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.xmlstream import ElementBase


class Bind(ElementBase):

    """
    """

    name = 'bind'
    namespace = 'urn:ietf:params:xml:ns:xmpp-bind'
    interfaces = {'resource', 'jid'}
    sub_interfaces = interfaces
    plugin_attrib = 'bind'




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2010 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins.base import register_plugin

from slixmpp.plugins.xep_0030 import stanza
from slixmpp.plugins.xep_0030.stanza import DiscoInfo, DiscoItems
from slixmpp.plugins.xep_0030.static import StaticDisco
from slixmpp.plugins.xep_0030.disco import XEP_0030


register_plugin(XEP_0030)

__all__ = ['DiscoInfo', 'DiscoItems', 'StaticDisco', 'XEP_0030', 'stanza']



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2020 Mathieu Pasquet <mathieui@mathieui.net>
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins.base import register_plugin
from slixmpp.plugins.xep_0428.stanza import (
    Fallback, FallbackBody, FallbackSubject
)
from slixmpp.plugins.xep_0428.fallback import XEP_0428

register_plugin(XEP_0428)

__all__ = ['Fallback', 'FallbackBody', 'FallbackSubject', 'XEP_0428']



from slixmpp import register_stanza_plugin
from slixmpp.plugins.xep_0402.stanza import Extensions
from slixmpp.xmlstream import ElementBase

NS = "urn:xmpp:bookmarks-pinning:0"


class Pinned(ElementBase):
    """
    Pinned bookmark element


    To enable it on a Conference element, use enable() like this:

    .. code-block::python

        # C being a Conference element
        C['extensions'].enable('pinned')

    Which will add the <pinned> element to the <extensions> element.
    """
    namespace = NS
    name = "pinned"
    plugin_attrib = "pinned"
    interfaces = {"pinned"}
    bool_interfaces = {"pinned"}
    is_extension = True


def register_plugin():
    register_stanza_plugin(Extensions, Pinned)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2010  Nathanael C. Fritz
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.jid import JID
from slixmpp.xmlstream.stanzabase import StanzaBase, ElementBase, ET
from slixmpp.xmlstream.stanzabase import register_stanza_plugin
from slixmpp.xmlstream.tostring import tostring, highlight
from slixmpp.xmlstream.xmlstream import XMLStream, RESPONSE_TIMEOUT

__all__ = ['JID', 'StanzaBase', 'ElementBase',
           'ET', 'tostring', 'highlight', 'XMLStream',
           'RESPONSE_TIMEOUT', 'register_stanza_plugin']




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2010 Nathanael C. Fritz
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.xmlstream import ElementBase


class Ping(ElementBase):

    """
    Given that XMPP is based on TCP connections, it is possible for the
    underlying connection to be terminated without the application's
    awareness. Ping stanzas provide an alternative to whitespace based
    keepalive methods for detecting lost connections.

    Example ping stanza:
    ::

        <iq type="get">
          <ping xmlns="urn:xmpp:ping" />
        </iq>
    """

    name = 'ping'
    namespace = 'urn:xmpp:ping'
    plugin_attrib = 'ping'
    interfaces = set()



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins.base import register_plugin

from slixmpp.plugins.xep_0221 import stanza
from slixmpp.plugins.xep_0221.stanza import Media, URI
from slixmpp.plugins.xep_0221.media import XEP_0221


register_plugin(XEP_0221)

__all__ = ['stanza', 'Media', 'URI', 'XEP_0221']




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2011 Nathanael C. Fritz, Dalek
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import logging

from slixmpp import Message, JID
from slixmpp.plugins import BasePlugin
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.xmlstream.handler import Callback
from slixmpp.xmlstream.matcher import StanzaPath
from slixmpp.plugins.xep_0249 import Invite, stanza


log = logging.getLogger(__name__)


class XEP_0249(BasePlugin):

    """
    XEP-0249: Direct MUC Invitations
    """

    name = 'xep_0249'
    description = 'XEP-0249: Direct MUC Invitations'
    dependencies = {'xep_0030'}
    stanza = stanza

    def plugin_init(self):
        self.xmpp.register_handler(
                Callback('Direct MUC Invitations',
                         StanzaPath('message/groupchat_invite'),
                         self._handle_invite))

        register_stanza_plugin(Message, Invite)

    def plugin_end(self):
        self.xmpp['xep_0030'].del_feature(feature=Invite.namespace)
        self.xmpp.remove_handler('Direct MUC Invitations')

    def session_bind(self, jid):
        self.xmpp['xep_0030'].add_feature(Invite.namespace)

    def _handle_invite(self, msg: Message):
        """
        Raise an event for all invitations received.
        """
        log.debug("Received direct muc invitation from %s to room %s",
                  msg['from'], msg['groupchat_invite']['jid'])

        self.xmpp.event('groupchat_direct_invite', msg)

    def send_invitation(self, jid: JID, roomjid: JID,
                        password: str | None = None,
                        reason: str | None = None, *,
                        mfrom: JID | None = None):
        """
        Send a direct MUC invitation to an XMPP entity.

        :param JID jid: The JID of the entity that will receive
                        the invitation
        :param JID roomjid: the address of the groupchat room to be joined
        :param str password: a password needed for entry into a
                             password-protected room (OPTIONAL).
        :param str reason: a human-readable purpose for the invitation
                            (OPTIONAL).
        """

        msg = self.xmpp.Message()
        msg['to'] = jid
        if mfrom is not None:
            msg['from'] = mfrom
        msg['groupchat_invite']['jid'] = roomjid
        if password is not None:
            msg['groupchat_invite']['password'] = password
        if reason is not None:
            msg['groupchat_invite']['reason'] = reason

        return msg.send()



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2025 Mathieu Pasquet
# This file is part of Slixmpp.
# See the file LICENSE for copying permission

from slixmpp import Message
from slixmpp.jid import JID
from slixmpp.xmlstream import ElementBase, register_stanza_plugin

NS = 'urn:xmpp:call-invites:0'


class Jingle(ElementBase):
    name = 'jingle'
    namespace = NS
    plugin_attrib = 'jingle'
    plugin_multi_attrib = 'jingles'
    interfaces = {'sid', 'jid'}

    def set_jid(self, value: JID) -> None:
        if not isinstance(value, JID):
            try:
                value = JID(value)
            except ValueError:
                raise ValueError('"jid" must be a valid JID object')
        self.xml.attrib['jid'] = value.full

    def get_jid(self) -> JID | None:
        try:
            return JID(self.xml.attrib.get('jid', ''))
        except ValueError:
            return None


class External(ElementBase):
    name = 'external'
    namespace = NS
    plugin_attrib = 'external'
    plugin_multi_attrib = 'externals'
    interfaces = {'uri'}


class Invite(ElementBase):
    name = 'invite'
    namespace = NS
    plugin_attrib = 'call-invite'
    interfaces = {'video'}

    def get_methods(self) -> tuple[list[Jingle], list[External]]:
        return (self['jingles'], self['externals'])

    def set_video(self, value: bool) -> None:
        if not isinstance(value, bool):
            raise ValueError(f'Invalid value for the video attribute: {value}')
        self.xml.attrib['video'] = str(value).lower()

    def get_video(self) -> bool:
        vid = self.xml.attrib.get('video', 'false').lower()
        return vid == 'true'


class Retract(ElementBase):
    name = 'retract'
    namespace = NS
    plugin_attrib = 'call-retract'
    interfaces = {'id'}


class Accept(ElementBase):
    name = 'accept'
    namespace = NS
    plugin_attrib = 'call-accept'
    interfaces = {'id'}


class Reject(ElementBase):
    name = 'reject'
    namespace = NS
    plugin_attrib = 'call-reject'
    interfaces = {'id'}


class Left(ElementBase):
    name = 'left'
    namespace = NS
    plugin_attrib = 'call-left'
    interfaces = {'id'}


def register_plugins() -> None:
    register_stanza_plugin(Message, Invite)
    register_stanza_plugin(Message, Retract)
    register_stanza_plugin(Message, Accept)
    register_stanza_plugin(Message, Reject)
    register_stanza_plugin(Message, Left)

    register_stanza_plugin(Invite, Jingle, iterable=True)
    register_stanza_plugin(Invite, External, iterable=True)

    register_stanza_plugin(Accept, Jingle)
    register_stanza_plugin(Accept, External)



from slixmpp.plugins.base import register_plugin

from . import stanza
from .privilege import XEP_0356
from .stanza import Perm, Privilege

register_plugin(XEP_0356)

__all__ = ['stanza', 'XEP_0356', 'Perm', 'Privilege', 'XEP_0356']



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permissio
from slixmpp.plugins.base import register_plugin

from slixmpp.plugins.xep_0280.stanza import ReceivedCarbon, SentCarbon
from slixmpp.plugins.xep_0280.stanza import PrivateCarbon
from slixmpp.plugins.xep_0280.stanza import CarbonEnable, CarbonDisable
from slixmpp.plugins.xep_0280.carbons import XEP_0280


register_plugin(XEP_0280)

__all__ = ['ReceivedCarbon', 'SentCarbon', 'PrivateCarbon', 'CarbonEnable',
           'CarbonDisable', 'XEP_0280']




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import logging
from asyncio import Future
from datetime import datetime, timedelta

from slixmpp.plugins import BasePlugin
from slixmpp import JID
from slixmpp.stanza import Iq
from slixmpp.exceptions import XMPPError
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.xmlstream.handler import CoroutineCallback
from slixmpp.xmlstream.matcher import StanzaPath
from slixmpp.plugins.xep_0012 import stanza, LastActivity


log = logging.getLogger(__name__)


class XEP_0012(BasePlugin):

    """
    XEP-0012 Last Activity
    """

    name = 'xep_0012'
    description = 'XEP-0012: Last Activity'
    dependencies = {'xep_0030'}
    stanza = stanza

    def plugin_init(self):
        register_stanza_plugin(Iq, LastActivity)

        self._last_activities = {}

        self.xmpp.register_handler(
            CoroutineCallback('Last Activity',
                 StanzaPath('iq@type=get/last_activity'),
                 self._handle_get_last_activity))

        self.api.register(self._default_get_last_activity,
                'get_last_activity',
                default=True)
        self.api.register(self._default_set_last_activity,
                'set_last_activity',
                default=True)
        self.api.register(self._default_del_last_activity,
                'del_last_activity',
                default=True)

    def plugin_end(self):
        self.xmpp.remove_handler('Last Activity')
        self.xmpp['xep_0030'].del_feature(feature='jabber:iq:last')

    def session_bind(self, jid):
        self.xmpp['xep_0030'].add_feature('jabber:iq:last')

    def begin_idle(self, jid: JID | None = None, status: str | None = None) -> Future:
        """Reset the last activity for the given JID.

        .. versionchanged:: 1.8.0
            This function now returns a Future.

        :param status: Optional status.
        """
        return self.set_last_activity(jid, 0, status)

    def end_idle(self, jid: JID | None = None) -> Future:
        """Remove the last activity of a JID.

        .. versionchanged:: 1.8.0
            This function now returns a Future.
        """
        return self.del_last_activity(jid)

    def start_uptime(self, status: str | None = None) -> Future:
        """
        .. versionchanged:: 1.8.0
            This function now returns a Future.
        """
        return self.set_last_activity(None, 0, status)

    def set_last_activity(self, jid=None, seconds=None, status=None) -> Future:
        """Set last activity for a JID.

        .. versionchanged:: 1.8.0
            This function now returns a Future.
        """
        return self.api['set_last_activity'](jid, args={
            'seconds': seconds,
            'status': status
        })

    def del_last_activity(self, jid: JID) -> Future:
        """Remove the last activity of a JID.

        .. versionchanged:: 1.8.0
            This function now returns a Future.
        """
        return self.api['del_last_activity'](jid)

    def get_last_activity(self, jid: JID, local: bool = False,
                          ifrom: JID | None = None, **iqkwargs) -> Future:
        """Get last activity for a specific JID.

        :param local: Fetch the value from the local cache.
        """
        if jid is not None and not isinstance(jid, JID):
            jid = JID(jid)

        if self.xmpp.is_component:
            if jid.domain == self.xmpp.boundjid.domain:
                local = True
        else:
            if str(jid) == str(self.xmpp.boundjid):
                local = True
        jid = jid.full

        if local or jid in (None, ''):
            log.debug("Looking up local last activity data for %s", jid)
            return self.api['get_last_activity'](jid, None, ifrom, None)

        iq = self.xmpp.make_iq_get(ito=jid, ifrom=ifrom)
        iq.enable('last_activity')
        return iq.send(**iqkwargs)

    async def _handle_get_last_activity(self, iq: Iq):
        log.debug("Received last activity query from " + \
                  "<%s> to <%s>.", iq['from'], iq['to'])
        reply = await self.api['get_last_activity'](iq['to'], None, iq['from'], iq)
        reply.send()

    # =================================================================
    # Default in-memory implementations for storing last activity data.
    # =================================================================

    def _default_set_last_activity(self, jid: JID, node: str, ifrom: JID, data: dict):
        seconds = data.get('seconds', None)
        if seconds is None:
            seconds = 0

        status = data.get('status', None)
        if status is None:
            status = ''

        self._last_activities[jid] = {
            'seconds': datetime.now() - timedelta(seconds=seconds),
            'status': status}

    def _default_del_last_activity(self, jid: JID, node: str, ifrom: JID, data: dict):
        if jid in self._last_activities:
            del self._last_activities[jid]

    def _default_get_last_activity(self, jid: JID, node: str, ifrom: JID, iq: Iq) -> Iq:
        if not isinstance(iq, Iq):
            reply = self.xmpp.Iq()
        else:
            reply = iq.reply()

        if jid not in self._last_activities:
            raise XMPPError('service-unavailable')

        bare = JID(jid).bare

        if bare != self.xmpp.boundjid.bare:
            if bare in self.xmpp.roster[jid]:
                sub = self.xmpp.roster[jid][bare]['subscription']
                if sub not in ('from', 'both'):
                    raise XMPPError('forbidden')

        td = datetime.now() - self._last_activities[jid]['seconds']
        seconds = td.seconds + td.days * 24 * 3600
        status = self._last_activities[jid]['status']

        reply['last_activity']['seconds'] = seconds
        reply['last_activity']['status'] = status

        return reply



from slixmpp import register_stanza_plugin
from slixmpp.plugins.xep_0060.stanza import Item
from slixmpp.xmlstream import ElementBase
from slixmpp.plugins.xep_0359.stanza import StanzaID

NS = "urn:xmpp:mds:displayed:0"


class Displayed(ElementBase):
    """
    Displayed element.

    .. code-block:: xml

        <displayed xmlns='urn:xmpp:mds:displayed:0'>
          <stanza-id xmlns='urn:xmpp:sid:0'
                     id='ca21deaf-812c-48f1-8f16-339a674f2864'
                     by='example@conference.shakespeare.lit'/>
        </displayed>
    """
    namespace = NS
    name = "displayed"
    plugin_attrib = "displayed"


def register_plugin():
    register_stanza_plugin(Displayed, StanzaID)
    register_stanza_plugin(Item, Displayed)



# Slixmpp: The Slick XMPP Library
# Copyright © 2026 nicoco <nicoco@nicoco.fr>
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.

from slixmpp.plugins import BasePlugin

from . import stanza


class XEP_0463(BasePlugin):
    """
    XEP-0463: MUC Affilition Versioning

    This plugin does only one thing: it registers the stanza plugins.
    """

    name = "xep_0463"
    description = "XEP-0463: MUC Affilition Versioning"
    dependencies = {"xep_0045"}
    stanza = stanza

    def plugin_init(self) -> None:
        stanza.register_plugin()



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins.base import register_plugin

from slixmpp.plugins.xep_0257 import stanza
from slixmpp.plugins.xep_0257.stanza import Certs, AppendCert, CertItem
from slixmpp.plugins.xep_0257.stanza import DisableCert, RevokeCert
from slixmpp.plugins.xep_0257.client_cert_management import XEP_0257


register_plugin(XEP_0257)

__all__ = ['stanza', 'Certs', 'AppendCert', 'DisableCert', 'RevokeCert',
           'XEP_0257', 'CertItem']



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz,
# Emmanuel Gil Peyrot <linkmauve@linkmauve.fr>
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import logging
import hashlib

from asyncio import Future

from slixmpp import JID
from slixmpp.exceptions import XMPPError
from slixmpp.stanza import Iq, Message, Presence
from slixmpp.xmlstream.handler import CoroutineCallback
from slixmpp.xmlstream.matcher import StanzaPath
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.plugins.base import BasePlugin
from slixmpp.plugins.xep_0231 import BitsOfBinary


log = logging.getLogger(__name__)


class XEP_0231(BasePlugin):

    """
    XEP-0231 Bits of Binary
    """

    name = 'xep_0231'
    description = 'XEP-0231: Bits of Binary'
    dependencies = {'xep_0030'}

    def plugin_init(self):
        self._cids = {}

        register_stanza_plugin(Iq, BitsOfBinary)
        register_stanza_plugin(Message, BitsOfBinary)
        register_stanza_plugin(Presence, BitsOfBinary)

        self.xmpp.register_handler(
            CoroutineCallback('Bits of Binary - Iq',
                StanzaPath('iq/bob'),
                self._handle_bob_iq))

        self.xmpp.register_handler(
            CoroutineCallback('Bits of Binary - Message',
                StanzaPath('message/bob'),
                self._handle_bob))

        self.xmpp.register_handler(
            CoroutineCallback('Bits of Binary - Presence',
                StanzaPath('presence/bob'),
                self._handle_bob))

        self.api.register(self._get_bob, 'get_bob', default=True)
        self.api.register(self._set_bob, 'set_bob', default=True)
        self.api.register(self._del_bob, 'del_bob', default=True)

    def plugin_end(self):
        self.xmpp['xep_0030'].del_feature(feature='urn:xmpp:bob')
        self.xmpp.remove_handler('Bits of Binary - Iq')
        self.xmpp.remove_handler('Bits of Binary - Message')
        self.xmpp.remove_handler('Bits of Binary - Presence')

    def session_bind(self, jid):
        self.xmpp['xep_0030'].add_feature('urn:xmpp:bob')

    async def set_bob(self, data: bytes, mtype: str, cid: str | None = None,
                      max_age: int | None = None) -> str:
        """Register a blob of binary data as a BOB.

        .. versionchanged:: 1.8.0
            If ``max_age`` is specified, the registered data will be destroyed
            after that time.
            This function is now a coroutine.

        :param data: Data to register.
        :param mtype: Mime Type of the data (e.g. ``image/jpeg``).
        :param cid: Content-ID (will be auto-generated if left out).
        :param max_age: Duration of content availability.
        :returns: The cid value.
        """
        if cid is None:
            cid = 'sha1+%s@bob.xmpp.org' % hashlib.sha1(data).hexdigest()

        bob = BitsOfBinary()
        bob['data'] = data
        bob['type'] = mtype
        bob['cid'] = cid
        if max_age is not None:
            bob['max_age'] = max_age

        await self.api['set_bob'](args=bob)
        # Schedule destruction of the data
        if max_age is not None and max_age > 0:
            self.xmpp.loop.call_later(max_age, self.del_bob,  cid)
        return cid

    async def get_bob(self, jid: JID | None = None, cid: str | None = None,
                      cached: bool = True, ifrom: JID | None = None,
                      **iqkwargs) -> Iq:
        """Get a BOB.

        .. versionchanged:: 1.8.0
            Results not in cache do not raise an error when ``cached`` is True.
            This function is now a coroutine.

        :param jid: JID to fetch the BOB from.
        :param cid: Content ID (actually required).
        :param cached: To fetch the BOB from the local cache first (from CID only)
        """
        if cached:
            data = await self.api['get_bob'](None, None, ifrom, args=cid)
            if data is not None:
                if not isinstance(data, Iq):
                    iq = self.xmpp.Iq()
                    iq.append(data)
                    return iq
                return data

        iq = self.xmpp.make_iq_get(ito=jid, ifrom=ifrom)
        iq['bob']['cid'] = cid
        return await iq.send(**iqkwargs)

    def del_bob(self, cid: str) -> Future:
        """Delete a stored BoB.

        .. versionchanged:: 1.8.0
            This function now returns a Future.
        """
        return self.api['del_bob'](args=cid)

    async def _handle_bob_iq(self, iq: Iq):
        cid = iq['bob']['cid']

        if iq['type'] == 'result':
            await self.api['set_bob'](iq['from'], None, iq['to'], args=iq['bob'])
            self.xmpp.event('bob', iq)
        elif iq['type'] == 'get':
            data = await self.api['get_bob'](iq['to'], None, iq['from'], args=cid)

            if data is None:
                raise XMPPError(
                    "item-not-found",
                    f"Bits of binary '{cid}' is not available",
                )

            if isinstance(data, Iq):
                data['id'] = iq['id']
                data.send()
                return

            iq = iq.reply()
            iq.append(data)
            iq.send()

    async def _handle_bob(self, stanza):
        await self.api['set_bob'](
            stanza['from'], None,
            stanza['to'], args=stanza['bob']
        )
        self.xmpp.event('bob', stanza)

    # =================================================================

    def _set_bob(self, jid, node, ifrom, bob):
        self._cids[bob['cid']] = bob

    def _get_bob(self, jid, node, ifrom, cid):
        if cid in self._cids:
            return self._cids[cid]

    def _del_bob(self, jid, node, ifrom, cid):
        if cid in self._cids:
            del self._cids[cid]



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2020 "Maxime “pep” Buquet <pep@bouah.net>"
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins.base import register_plugin
from slixmpp.plugins.xep_0421.stanza import OccupantId
from slixmpp.plugins.xep_0421.occupant_id import XEP_0421

register_plugin(XEP_0421)

__all__ = ['OccupantId', 'XEP_0421']



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins import BasePlugin
from slixmpp.xmlstream import StanzaBase
from slixmpp.xmlstream.handler import Callback
from slixmpp.xmlstream.matcher import StanzaPath

from . import stanza


class XEP_0513(BasePlugin):
    """
    This plugin add supports for XEP-0513: Explicit Mentions

    """

    name = "xep_0513"
    description = "Explicit Mentions"
    dependencies = set()
    stanza = stanza

    def plugin_init(self) -> None:
        stanza.register_plugin()
        self.xmpp.register_handler(
            Callback(
                "Explicit Mentions", StanzaPath("message/mention"), self._handle_mention
            )
        )

    def plugin_end(self):
        self.xmpp.remove_handler("Explicit Mentions")

    def _handle_mention(self, msg: StanzaBase) -> None:
        self.xmpp.event("mention", msg)



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins.base import register_plugin

from slixmpp.plugins.xep_0118 import stanza
from slixmpp.plugins.xep_0118.stanza import UserTune
from slixmpp.plugins.xep_0118.user_tune import XEP_0118


register_plugin(XEP_0118)

__all__ = ['stanza', 'UserTune', 'XEP_0118']



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2025 Mathieu Pasquet
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.

from slixmpp import Iq
from slixmpp.plugins import BasePlugin
from . import stanza


class XEP_0494(BasePlugin):
    """
    XEP-0494: Client Access Management
    """

    name = "xep_0494"
    description = "XEP-0494: Client Access Management"
    dependencies = {}

    def plugin_init(self):
        stanza.register_plugins()

    async def get_clients(self, *, timeout: float | None = None,
                          **iqargs) -> list[stanza.Client]:
        """
        Return a list of clients that have accessed the account.

        :raises IqTimeout: If the request times out.
        :raises IqError: If the server answers with an error.
        """
        iq = self.xmpp.make_iq_get(**iqargs)
        iq.enable('list')
        iq_result = await iq.send(timeout=timeout)
        return list(iq_result['clients'])

    async def revoke(self, client_id: str, *, timeout: float | None = None,
                     **iqargs) -> Iq:
        """
        Revoke a specific client access by id.
        Revoking clients that have password access requires to change the
        password and will raise an error.

        :param client_id: id of the client to revoke.
        :raises IqTimeout: If the request times out.
        :raises IqError: If the server answers with an error.
        """
        iq = self.xmpp.make_iq_get(**iqargs)
        iq['revoke']['id'] = client_id
        return await iq.send(timeout=timeout)



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins.base import register_plugin

from slixmpp.plugins.xep_0047 import stanza
from slixmpp.plugins.xep_0047.stanza import Open, Close, Data
from slixmpp.plugins.xep_0047.stream import IBBytestream
from slixmpp.plugins.xep_0047.ibb import XEP_0047


register_plugin(XEP_0047)

__all__ = ['Open', 'Close', 'Data', 'IBBytestream', 'XEP_0047', 'stanza']



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2025 Mathieu Pasquet
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.

from datetime import datetime
from enum import Enum
from slixmpp import Iq, register_stanza_plugin, ElementBase
from slixmpp.plugins.xep_0082 import parse, format_datetime

NS = 'urn:xmpp:cam:0'


class ClientType(Enum):
    """
    Enum representing the ``type`` attribute of the ``<client/>`` element.
    """
    SESSION = 'session'
    ACCESS = 'access'


class PermissionEnum(Enum):
    """
    Enum representing the ``status`` of the ``<permission/>`` element, e.g:

    .. code-block:: xml

        <permission status="normal/>

    """
    UNRESTRICTED = 'unrestricted'
    NORMAL = 'normal'
    RESTRICTED = 'restricted'


class Revoke(ElementBase):
    """
    Element used to revoke a device.
    """
    namespace = NS
    name = 'revoke'
    plugin_attrib = 'revoke'
    interfaces = {'id'}


class List(ElementBase):
    """
    Element used to request a list of devices to the server.
    """
    namespace = NS
    name = 'list'
    plugin_attrib = 'list'
    interfaces = {}


class Clients(ElementBase):
    """List of clients.

    Iterate on it to get all substanzas.

    .. code-block::

        <clients xmlns="urn:xmpp:cam:0">
            …
        </clients>
    """
    namespace = NS
    name = 'clients'
    plugin_attrib = 'clients'


class Client(ElementBase):
    """Client element.

    .. code-block:: xml

        <client xmlns="urn:xmpp:cam:0" type="session" id="a" connected="true">
            …
        </client>

    Many of the substanzas defined are using overrides to access and set
    data from this element, e.g.:

    .. code-block:: python

        client = Client()
        client['user_agent'] = {'software': 'slixmpp', 'device': 'toto'}

    will actually create

    .. code-block:: xml

        <client xmlns="urn:xmpp:cam:0">
            <user-agent>
                <software>slixmpp</software>
                <device>toto</device>
            </user-agent>
        </client>

    """
    namespace = NS
    name = 'client'
    plugin_attrib = 'client'
    interfaces = {
        'connected', 'id', 'type', 'first_seen', 'last_seen',
        'permission', 'permission_extra',
    }

    def get_type(self) -> ClientType | None:
        type_ = self.xml.attrib.get('type', None)
        if type_ is not None:
            return ClientType(type_)
        return None

    def set_type(self, type_: str | ClientType) -> None:
        if isinstance(type_, ClientType):
            value = type_.value
        else:
            ClientType(type_)
            value = type_
        self.xml.attrib['type'] = value

    def get_connected(self) -> bool | None:
        connected = self.xml.attrib.get('connected', None)
        if connected is None or connected.lower() not in ('true', 'false'):
            return None
        return connected.lower() == 'true'

    def set_connected(self, connected: bool) -> None:
        self.xml.attrib['connected'] = str(connected).lower()

    def get_first_seen(self) -> datetime | None:
        try:
            parse(self._get_sub_text('first-seen'))
        except:
            return None

    def get_last_seen(self) -> datetime | None:
        try:
            parse(self._get_sub_text('last-seen'))
        except:
            return None

    def set_first_seen(self, time: str | datetime) -> None:
        if isinstance(time, str):
            time = parse(time)
        self._set_sub_text('first-seen', format_datetime(time))

    def set_last_seen(self, time: str | datetime) -> None:
        if isinstance(time, str):
            time = parse(time)
        self._set_sub_text('last-seen', format_datetime(time))

    def del_permission(self) -> None:
        found = self.xml.findall('{%s}permission' % NS)
        if found:
            for permission in found:
                self.xml.remove(permission)

    def del_user_agent(self) -> None:
        found = self.xml.findall('{%s}user-agent' % NS)
        if found:
            for agent in found:
                self.xml.remove(agent)


class UserAgent(ElementBase):
    namespace = NS
    name = 'user-agent'
    plugin_attrib = 'user_agent'
    interfaces = {'software', 'uri', 'device'}
    sub_interfaces = {'software', 'uri', 'device'}

    def as_dict(self) -> dict[str, str]:
        return {key: self[key] for key in self.sub_interfaces}


class Permission(ElementBase):
    """
    Permission element.
    """
    namespace = NS
    name = 'permission'
    plugin_attrib = 'permission'
    interfaces = {'permission', 'permission_extra'}
    overrides = {
        'set_permission', 'get_permission',
        'set_permission_extra', 'get_permission_extra', 'del_permission_extra',
    }

    def get_permission(self) -> PermissionEnum | None:
        return PermissionEnum(self.xml.attrib.get('status'))

    def set_permission(self, permission: str | PermissionEnum) -> None:
        if isinstance(permission, PermissionEnum):
            value = permission.value
        else:
            PermissionEnum(permission)
            value = permission
        self.xml.attrib['status'] = value

    def get_permission_extra(self) -> list[ElementBase]:
        return list(self)

    def set_permission_extra(self, extras: list[ElementBase]) -> None:
        self.clear()
        for extra in extras:
            self.append(extra)

    def del_permission_extra(self) -> None:
        self.clear()


class Auth(ElementBase):
    """
    Auth element.


    .. code-block:: xml

        <auth>
            <fast/>
            <custom-element xmlns="urn:custom"/>
        </auth>
    """
    namespace = NS
    name = 'auth'
    plugin_attrib = 'auth'
    interfaces = {'password', 'fast', 'all'}
    bool_interfaces = {'password', 'fast'}
    overrides = {'get_auths'}

    def get_auths(self) -> dict:
        return {
            'password': self['password'],
            'grant': 'grant' in self,
            'fast': self['fast'],
            'others': [sub for sub in self],
        }


class Grant(ElementBase):
    """``<grant/>`` element."""
    namespace = NS
    name = 'grant'
    plugin_attrib = 'grant'


def register_plugins() -> None:
    register_stanza_plugin(Iq, Clients)
    register_stanza_plugin(Iq, List)
    register_stanza_plugin(Iq, Revoke)
    register_stanza_plugin(Clients, Client, iterable=True)
    register_stanza_plugin(Client, UserAgent, overrides=True)
    register_stanza_plugin(Client, Permission, overrides=True)
    register_stanza_plugin(Client, Auth, overrides=True)
    register_stanza_plugin(Grant, Permission, overrides=True)
    register_stanza_plugin(Auth, Grant)



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins.base import register_plugin

from slixmpp.plugins.xep_0191.stanza import Block, Unblock, BlockList, BlockItem
from slixmpp.plugins.xep_0191.blocking import XEP_0191


register_plugin(XEP_0191)

__all__ = ['Block', 'Unblock', 'BlockList', 'BlockItem', 'XEP_0191']



import logging

from slixmpp import Message
from slixmpp.plugins import BasePlugin
from slixmpp.xmlstream.handler import Callback
from slixmpp.xmlstream.matcher import StanzaPath

from . import stanza

log = logging.getLogger(__name__)


class XEP_0449(BasePlugin):
    """
    XEP-0449: Stickers

    Raises an event on incoming stickers.
    Nothing related to sticker packs is supported (yet).
    """

    name = "xep_0449"
    description = "XEP-0449: Stickers"
    dependencies = {"xep_0449"}
    stanza = stanza

    def plugin_init(self) -> None:
        stanza.register()
        self.xmpp.register_handler(
            Callback(
                "Sticker received",
                StanzaPath("message/sticker"),
                self._handle_sticker,
            )
        )

    def plugin_end(self) -> None:
        self.xmpp.remove_handler("Sticker received")

    def _handle_sticker(self, msg: Message) -> None:
        self.xmpp.event("sticker", msg)



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2020 Mathieu Pasquet <mathieui@mathieui.net>
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from typing import (
    Any,
)

from datetime import datetime
from slixmpp import JID
from slixmpp.plugins import BasePlugin
from slixmpp.plugins.xep_0369 import stanza
from slixmpp.xmlstream.handler import Callback
from slixmpp.xmlstream.matcher import MatchXPath

try:
    from typing import TypedDict
    InfoType = TypedDict(
        'InfoType',
        {
            'Name': str,
            'Description': str,
            'Contact': list[JID] | None,
            'modified': datetime
        },
        total=False,
    )
except ImportError:
    # Placeholder until we drop python < 3.8
    InfoType = dict[str, Any]


BASE_NODES = [
    'urn:xmpp:mix:nodes:messages',
    'urn:xmpp:mix:nodes:participants',
    'urn:xmpp:mix:nodes:info',
]


class XEP_0369(BasePlugin):
    '''XEP-0369: MIX-CORE'''

    name = 'xep_0369'
    description = 'XEP-0369: MIX-CORE'
    dependencies = {'xep_0030', 'xep_0060', 'xep_0082', 'xep_0004'}
    stanza = stanza
    namespace = stanza.NS

    def plugin_init(self) -> None:
        stanza.register_plugins()
        self.xmpp.register_handler(
            Callback(
                "MIX message received",
                MatchXPath('{%s}message[@type="groupchat"]/{%s}mix' % (
                    self.xmpp.default_ns, self.namespace
                )),
                self._handle_mix_message,
            )
        )

    def _handle_mix_message(self, message):
        self.xmpp.event('mix_message', message)

    def session_bind(self, jid):
        self.xmpp.plugin['xep_0030'].add_feature(stanza.NS)
        self.xmpp.plugin['xep_0060'].map_node_event(
            'urn:xmpp:mix:nodes:participants',
            'mix_participant_info',
        )
        self.xmpp.plugin['xep_0060'].map_node_event(
            'urn:xmpp:mix:nodes:info',
            'mix_channel_info',
        )

    def plugin_end(self):
        self.xmpp.plugin['xep_0030'].del_feature(feature=stanza.NS)
        node_map = self.xmpp.plugin['xep_0060'].node_event_map
        if 'urn:xmpp:mix:nodes:info' in node_map:
            del node_map['urn:xmpp:mix:nodes:info']
        if 'urn:xmpp:mix:nodes:participants' in node_map:
            del node_map['urn:xmpp:mix:nodes:participants']

    async def get_channel_info(self, channel: JID) -> InfoType:
        """"
        Get the contents of the channel info node.

        :param channel: The MIX channel
        :returns: a dict containing the last modified time and form contents
                  (Name, Description, Contact per the spec, YMMV)
        """
        info = await self.xmpp['xep_0060'].get_items(channel, 'urn:xmpp:mix:nodes:info')
        for item in info['pubsub']['items']:
            time = item['id']
            fields = item['form'].get_values()
            del fields['FORM_TYPE']
            fields['modified'] = self.xmpp['xep_0082'].parse(time)
            contact = fields.get('Contact')
            if contact:
                if isinstance(contact, str):
                    contact = [contact]
                elif isinstance(contact, list):
                    contact = [JID(cont) for cont in contact]
                fields['Contact'] = contact
            return fields

    async def join_channel(self, channel: JID, nick: str, subscribe: set[str] | None = None, *,
                           ifrom: JID | None = None, **iqkwargs) -> set[str]:
        """
        Join a MIX channel.

        :param JID channel: JID of the MIX channel
        :param str nick: Desired nickname on that channel
        :param set[str] subscribe: Set of notes to subscribe to when joining.
            If empty, all nodes will be subscribed by default.

        :rtype: set[str]
        :return: The nodes that failed to subscribe, if any
        """
        if not subscribe:
            subscribe = set(BASE_NODES)
        iq = self.xmpp.make_iq_set(ito=channel, ifrom=ifrom)
        iq['mix_join']['nick'] = nick
        for node in subscribe:
            sub = stanza.Subscribe()
            sub['node'] = node
            iq['mix_join']['subscribe'].append(sub)
        result = await iq.send(**iqkwargs)
        result_nodes = {sub['node'] for sub in result['mix_join']}
        return result_nodes.difference(subscribe)

    async def update_subscription(self, channel: JID,
                                  subscribe: set[str] | None = None,
                                  unsubscribe: set[str] | None = None, *,
                                  ifrom: JID | None = None, **iqkwargs) -> tuple[set[str], set[str]]:
        """
        Update a MIX channel subscription.

        :param JID channel: JID of the MIX channel
        :param set[str] subscribe: Set of notes to subscribe to additionally.
        :param set[str] unsubscribe: Set of notes to unsubscribe from.
        :rtype: tuple[set[str], set[str]]
        :return: A tuple containing the set of nodes that failed to subscribe
            and the set of nodes that failed to unsubscribe.
        """
        if not subscribe and not unsubscribe:
            raise ValueError("No nodes were provided.")
        unsubscribe = unsubscribe or set()
        subscribe = subscribe or set()
        iq = self.xmpp.make_iq_set(ito=channel, ifrom=ifrom)
        iq.enable('mix_updatesub')
        for node in subscribe:
            sub = stanza.Subscribe()
            sub['node'] = node
            iq['mix_updatesub'].append(sub)
        for node in unsubscribe:
            unsub = stanza.Unsubscribe()
            unsub['node'] = node
            iq['mix_updatesub'].append(unsub)
        result = await iq.send(**iqkwargs)
        for item in result['mix_updatesub']:
            if isinstance(item, stanza.Subscribe):
                subscribe.discard(item['node'])
            elif isinstance(item, stanza.Unsubscribe):
                unsubscribe.discard(item['node'])
        return (subscribe, unsubscribe)

    async def leave_channel(self, channel: JID, *,
                            ifrom: JID | None = None, **iqkwargs) -> None:
        """"
        Leave a MIX channel
        :param JID channel: JID of the channel to leave
        """
        iq = self.xmpp.make_iq_set(ito=channel, ifrom=ifrom)
        iq.enable('mix_leave')
        await iq.send(**iqkwargs)

    async def set_nick(self, channel: JID, nick: str, *,
                       ifrom: JID | None = None, **iqkwargs) -> str:
        """
        Set your nick on a channel. The returned nick MAY be different
        from the one provided, depending on service configuration.
        :param JID channel: MIX channel JID
        :param str nick: desired nick
        :rtype: str
        :return: The nick saved on the MIX channel
        """

        iq = self.xmpp.make_iq_set(ito=channel, ifrom=ifrom)
        iq['mix_setnick']['nick'] = nick
        result = await iq.send(**iqkwargs)
        result_nick = result['mix_setnick']['nick']
        return result_nick

    async def can_create_channel(self, service: JID) -> bool:
        """
        Check if the current user can create a channel on the MIX service

        :param JID service: MIX service jid
        :rtype: bool
        """
        results_stanza = await self.xmpp['xep_0030'].get_info(service.server)
        features = results_stanza['disco_info']['features']
        return 'urn:xmpp:mix:core:1#create-channel' in features

    async def create_channel(self, service: JID, channel: str | None = None, *,
                             ifrom: JID | None = None, **iqkwargs) -> str:
        """
        Create a MIX channel.

        :param JID service: MIX service JID
        :param str | None channel: Channel name (or leave empty to let
            the service generate it)
        :returns: The channel name, as created by the service
        """
        if '#' in channel:
            raise ValueError("A channel name cannot contain hashes")
        iq = self.xmpp.make_iq_set(ito=service.server, ifrom=ifrom)
        iq.enable('mix_create')
        if channel is not None:
            iq['mix_create']['channel'] = channel
        result = await iq.send(**iqkwargs)
        return result['mix_create']['channel']

    async def destroy_channel(self, channel: JID, *,
                              ifrom: JID | None = None, **iqkwargs):
        """
        Destroy a MIX channel.
        :param JID channel: MIX channelJID
        """
        iq = self.xmpp.make_iq_set(ito=channel.server, ifrom=ifrom)
        iq['mix_destroy'] = channel.user
        await iq.send(**iqkwargs)

    async def list_mix_nodes(self, channel: JID,
                             ifrom: JID | None = None, **discokwargs) -> set[str]:
        """
        List mix nodes for a channel.

        :param JID channel: The MIX channel
        :returns: List of nodes available
        """
        result = await self.xmpp['xep_0030'].get_items(
            channel,
            node='mix',
            ifrom=ifrom,
            **discokwargs,
        )
        nodes = set()
        for item in result['disco_items']:
            nodes.add(item['node'])
        return nodes

    async def list_participants(self, channel: JID, *,
                                ifrom: JID | None = None, **pubsubkwargs) -> list[tuple[str, str, JID | None]]:
        """
        List the participants of a MIX channel
        :param JID channel: The MIX channel

        :returns: A list of tuples containing the participant id, nick, and jid (if available)
        """
        info = await self.xmpp['xep_0060'].get_items(
            channel,
            'urn:xmpp:mix:nodes:participants',
            ifrom=ifrom,
            **pubsubkwargs
        )
        participants = list()
        for item in info['pubsub']['items']:
            identifier = item['id']
            nick = item['mix_participant']['nick']
            jid = item['mix_participant']['jid'] or None
            participants.append(
                (identifier, nick, jid),
            )
        return participants

    async def list_channels(self, service: JID, *,
                            ifrom: JID | None =None, **discokwargs) -> list[tuple[JID, str]]:
        """
        List the channels on a MIX service

        :param JID service: MIX service JID
        :returns: A list of channels with their JID and name
        """
        results_stanza = await self.xmpp['xep_0030'].get_items(
            service.server,
            ifrom=ifrom,
            **discokwargs,
        )
        results = []
        for result in results_stanza['disco_items']:
            results.append((result['jid'], result['name']))
        return results



from slixmpp.xmlstream import ElementBase

NAMESPACE = "urn:xmpp:sfs:0"


class StatelessFileSharing(ElementBase):
    name = "file-sharing"
    plugin_attrib = "sfs"
    namespace = NAMESPACE
    interfaces = {"disposition"}


class Sources(ElementBase):
    name = plugin_attrib = "sources"
    namespace = NAMESPACE


class UrlData(ElementBase):
    name = plugin_attrib = "url-data"
    namespace = "http://jabber.org/protocol/url-data"
    interfaces = {"target"}




# slixmpp.util
# ~~~~~~~~~~~~~~
# Part of Slixmpp: The Slick XMPP Library
# :copyright: (c) 2012 Nathanael C. Fritz, Lance J.T. Stout
# :license: MIT, see LICENSE for more details

from slixmpp.util.misc_ops import bytes, unicode, hashes, hash, \
                                    num_to_bytes, bytes_to_num, quote, \
                                    XOR
from slixmpp.util.cache import MemoryCache, MemoryPerJidCache, \
                               FileSystemCache, FileSystemPerJidCache




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission
import logging

from asyncio import Future
from collections.abc import AsyncGenerator
from datetime import datetime
from typing import (
    Any,
    Awaitable,
    Callable,
)

from slixmpp import JID
from slixmpp.stanza import Message, Iq
from slixmpp.xmlstream.handler import Collector
from slixmpp.xmlstream.matcher import MatchXMLMask
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.plugins import BasePlugin
from slixmpp.plugins.xep_0313 import stanza
from slixmpp.plugins.xep_0004.stanza import Form


log = logging.getLogger(__name__)


class XEP_0313(BasePlugin):

    """
    XEP-0313 Message Archive Management
    """

    name = 'xep_0313'
    description = 'XEP-0313: Message Archive Management'
    dependencies = {
        'xep_0004', 'xep_0030', 'xep_0050', 'xep_0059', 'xep_0297'
    }
    stanza = stanza

    def plugin_init(self):
        register_stanza_plugin(stanza.MAM, Form)
        register_stanza_plugin(Iq, stanza.MAM)
        register_stanza_plugin(Message, stanza.Result)
        register_stanza_plugin(Iq, stanza.Fin)
        register_stanza_plugin(
            stanza.Result,
            self.xmpp['xep_0297'].stanza.Forwarded
        )
        register_stanza_plugin(stanza.MAM, self.xmpp['xep_0059'].stanza.Set)
        register_stanza_plugin(stanza.Fin, self.xmpp['xep_0059'].stanza.Set)
        register_stanza_plugin(Iq, stanza.Metadata)
        register_stanza_plugin(stanza.Metadata, stanza.Start)
        register_stanza_plugin(stanza.Metadata, stanza.End)

    def retrieve(
            self,
            jid: JID | None = None,
            start: datetime | None = None,
            end: datetime | None = None,
            with_jid: JID | None = None,
            ifrom: JID | None = None,
            reverse: bool = False,
            timeout: int | None = None,
            callback: Callable[[Iq], None] | None = None,
            iterator: bool = False,
            rsm: dict[str, Any] | None = None
    ) -> Awaitable[Iq]:
        """
        Send a MAM query and retrieve the results.

        :param JID jid: Entity holding the MAM records
        :param datetime start,end: MAM query temporal boundaries
        :param JID with_jid: Filter results on this JID
        :param JID ifrom: To change the from address of the query
        :param bool reverse: Get the results in reverse order
        :param int timeout: IQ timeout
        :param func callback: Custom callback for handling results
        :param bool iterator: Use RSM and iterate over a paginated query
        :param dict rsm: RSM custom options
        """
        iq, stanza_mask = self._pre_mam_retrieve(
            jid, start, end, with_jid, ifrom
        )
        query_id = iq['id']
        amount = 10
        if rsm:
            for key, value in rsm.items():
                iq['mam']['rsm'][key] = str(value)
                if key == 'max':
                    amount = value
        cb_data = {}

        xml_mask = str(stanza_mask)

        def pre_cb(query: Iq) -> None:
            stanza_mask['mam_result']['queryid'] = query['id']
            xml_mask = str(stanza_mask)
            query['mam']['queryid'] = query['id']
            collector = Collector(
                'MAM_Results_%s' % query_id,
                MatchXMLMask(xml_mask))
            self.xmpp.register_handler(collector)
            cb_data['collector'] = collector

        def post_cb(result: Iq) -> None:
            results = cb_data['collector'].stop()
            if result['type'] == 'result':
                result['mam']['results'] = results
                result['mam_fin']['results'] = results

        if iterator:
            return self.xmpp['xep_0059'].iterate(
                iq, 'mam', 'results', amount=amount,
                reverse=reverse, recv_interface='mam_fin',
                pre_cb=pre_cb, post_cb=post_cb
            )

        collector = Collector(
            'MAM_Results_%s' % query_id,
            MatchXMLMask(xml_mask))
        self.xmpp.register_handler(collector)

        def wrapped_cb(iq: Iq) -> None:
            results = collector.stop()
            if iq['type'] == 'result':
                iq['mam']['results'] = results
            if callback:
                callback(iq)

        return iq.send(timeout=timeout, callback=wrapped_cb)

    async def iterate(
            self,
            jid: JID | None = None,
            start: datetime | None = None,
            end: datetime | None = None,
            with_jid: JID | None = None,
            ifrom: JID | None = None,
            reverse: bool = False,
            rsm: dict[str, Any] | None = None,
            total: int | None = None,
    ) -> AsyncGenerator[Message, None]:
        """
        Iterate over each message of MAM query.

        .. versionadded:: 1.8.0

        :param jid: Entity holding the MAM records
        :param start: MAM query start time
        :param end: MAM query end time
        :param with_jid: Filter results on this JID
        :param ifrom: To change the from address of the query
        :param reverse: Get the results in reverse order
        :param rsm: RSM custom options
        :param total: A number of messages received after which the query
                      should stop.
        """
        iq, stanza_mask = self._pre_mam_retrieve(
            jid, start, end, with_jid, ifrom
        )
        query_id = iq['id']
        amount = 10

        if rsm:
            for key, value in rsm.items():
                iq['mam']['rsm'][key] = str(value)
                if key == 'max':
                    amount = value
        cb_data = {}

        def pre_cb(query: Iq) -> None:
            stanza_mask['mam_result']['queryid'] = query['id']
            xml_mask = str(stanza_mask)
            query['mam']['queryid'] = query['id']
            collector = Collector(
                'MAM_Results_%s' % query_id,
                MatchXMLMask(xml_mask))
            self.xmpp.register_handler(collector)
            cb_data['collector'] = collector

        def post_cb(result: Iq) -> None:
            results = cb_data['collector'].stop()
            if result['type'] == 'result':
                result['mam']['results'] = results
                result['mam_fin']['results'] = results

        iterator = self.xmpp['xep_0059'].iterate(
            iq, 'mam', 'results', amount=amount,
            reverse=reverse, recv_interface='mam_fin',
            pre_cb=pre_cb, post_cb=post_cb
        )
        recv_count = 0

        async for page in iterator:
            messages = [message for message in page['mam']['results']]
            if reverse:
                messages.reverse()
            for message in messages:
                yield message
                recv_count += 1
                if total is not None and recv_count >= total:
                    break
            if total is not None and recv_count >= total:
                break

    def _pre_mam_retrieve(
            self,
            jid: JID | None = None,
            start: datetime | None = None,
            end: datetime | None = None,
            with_jid: JID | None = None,
            ifrom: JID | None = None,
    ) -> tuple[Iq, Message]:
        """Build the IQ and stanza mask for MAM results
        """
        iq = self.xmpp.make_iq_set(ito=jid, ifrom=ifrom)
        query_id = iq['id']
        iq['mam']['queryid'] = query_id
        iq['mam']['start'] = start
        iq['mam']['end'] = end
        iq['mam']['with'] = with_jid

        stanza_mask = self.xmpp.Message()

        auto_origin = stanza_mask.xml.find('{urn:xmpp:sid:0}origin-id')
        if auto_origin is not None:
            stanza_mask.xml.remove(auto_origin)
        del stanza_mask['id']
        del stanza_mask['lang']
        stanza_mask['from'] = jid
        stanza_mask['mam_result']['queryid'] = query_id

        return (iq, stanza_mask)

    async def get_fields(self, jid: JID | None = None, **iqkwargs) -> Form:
        """Get MAM query fields.

        .. versionadded:: 1.8.0

        :param jid: JID to retrieve the policy from.
        :return: The Form of allowed options
        """
        ifrom = iqkwargs.pop('ifrom', None)
        iq = self.xmpp.make_iq_get(ito=jid, ifrom=ifrom)
        iq.enable('mam')
        result = await iq.send(**iqkwargs)
        return result['mam']['form']

    async def get_configuration_commands(self, jid: JID | str | None,
                                         **discokwargs) -> Future[Iq]:
        """Get the list of MAM advanced configuration commands.

        .. versionchanged:: 1.8.0

        :param jid: JID to get the commands from.
        """
        if jid is None:
            jid = self.xmpp.boundjid.bare
        return await self.xmpp['xep_0030'].get_items(
            jid=jid,
            node='urn:xmpp:mam#configure',
            **discokwargs
        )

    def get_archive_metadata(self, jid: JID | None = None,
                             **iqkwargs) -> Future[Iq]:
        """Get the archive metadata from a JID.

        :param jid: JID to get the metadata from.
        """
        ifrom = iqkwargs.pop('ifrom', None)
        iq = self.xmpp.make_iq_get(ito=jid, ifrom=ifrom)
        iq.enable('mam_metadata')
        return iq.send(**iqkwargs)



from datetime import datetime

from slixmpp.plugins.xep_0082 import format_datetime, parse
from slixmpp.plugins.xep_0300 import Hash
from slixmpp.plugins.xep_0264.stanza import Thumbnail
from slixmpp.xmlstream import ElementBase, register_stanza_plugin

NS = "urn:xmpp:file:metadata:0"


class File(ElementBase):
    """
    File metadata element

    .. code-block:: xml

        <file xmlns='urn:xmpp:file:metadata:0'>
            <media-type>text/plain</media-type>
            <name>test.txt</name>
            <date>2015-07-26T21:46:00+01:00</date>
            <size>6144</size>
            <hash xmlns='urn:xmpp:hashes:2'
                  algo='sha-1'>w0mcJylzCn+AfvuGdqkty2+KP48=</hash>
        </file>
    """
    name = "file"
    namespace = NS
    plugin_attrib = "file"
    interfaces = sub_interfaces = {
        "media-type",
        "name",
        "date",
        "size",
        "desc",
        "width",
        "height",
        "length"
    }

    def set_width(self, width: int):
        self.__set_if_positive("width", width)

    def get_width(self) -> int | None:
        return _positive_int_or_none(self._get_sub_text("width"))

    def set_height(self, height: int):
        self.__set_if_positive("height", height)

    def get_height(self) -> int | None:
        return _positive_int_or_none(self._get_sub_text("height"))

    def set_length(self, length: int):
        self.__set_if_positive("length", length)

    def get_length(self) -> int | None:
        return _positive_int_or_none(self._get_sub_text("length"))

    def set_size(self, size: int):
        self.__set_if_positive("size", size)

    def get_size(self) -> int | None:
        return _positive_int_or_none(self._get_sub_text("size"))

    def get_date(self) -> datetime | None:
        try:
            return parse(self._get_sub_text("date"))
        except ValueError:
            return

    def set_date(self, stamp: datetime):
        try:
            self._set_sub_text("date", format_datetime(stamp))
        except ValueError:
            pass

    def __set_if_positive(self, key: str, value: int):
        if value <= 0:
            raise ValueError(f"Invalid value for element {key}: {value}")
        self._set_sub_text(key, str(value))


def _positive_int_or_none(v: str) -> int | None:
    try:
        return int(v)
    except ValueError:
        return None


def register_plugins():
    register_stanza_plugin(File, Hash)
    register_stanza_plugin(File, Thumbnail)



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2011 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins.base import register_plugin

from slixmpp.plugins.xep_0086.stanza import LegacyError
from slixmpp.plugins.xep_0086.legacy_error import XEP_0086


register_plugin(XEP_0086)

__all__ = ['LegacyError', 'XEP_0086']




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2018 Maxime “pep” Buquet
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import json
from slixmpp.xmlstream import ElementBase


class JSON_Container(ElementBase):
    name = 'json'
    plugin_attrib = 'json'
    namespace = 'urn:xmpp:json:0'
    interfaces = {'value'}

    def get_value(self):
        return json.loads(self.xml.text)

    def set_value(self, value):
        if not isinstance(value, str):
            value = json.dumps(value)
        self.xml.text = value

    def del_value(self):
        self.xml.text = ''




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.xmlstream import ElementBase


class BlockList(ElementBase):
    name = 'blocklist'
    namespace = 'urn:xmpp:blocking'
    plugin_attrib = 'blocklist'


class Block(BlockList):
    name = 'block'
    plugin_attrib = 'block'


class Unblock(BlockList):
    name = 'unblock'
    plugin_attrib = 'unblock'


class BlockItem(ElementBase):
    name = 'item'
    namespace = 'urn:xmpp:blocking'
    plugin_attrib = 'item'
    plugin_multi_attrib = 'items'
    interfaces = {'jid'}
